"""Error types for ``pane`` library."""

from __future__ import annotations

import abc
from io import StringIO
import sys
import traceback
import dataclasses
import typing as t


class ParseInterrupt(Exception):
    """
    Raised by [`Converter`][pane.converters.Converter]s to indicate that a given parsing path has failed
    (without materializing a detailed error message).
    """
    ...


class UnsupportedAnnotation(Exception):
    """
    Raised when a given [`t.Annotated`][typing.Annotated] isn't understood by `pane`.
    """
    def __init__(self, obj: t.Any):
        self.obj: t.Any = obj

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.obj!r})"

    def __str__(self) -> str:
        return f"Unsupported annotation: {self.obj!r}"


class ConvertError(Exception):
    """
    `pane` conversion error.

    `self.tree` contains a detailed error tree, and `str(self)`
    is a human-friendly representation of the same.
    """
    def __init__(self, tree: ErrorNode):
        self.tree: ErrorNode = tree

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tree!r})"

    def __str__(self) -> str:
        return str(self.tree)


class ErrorNode(abc.ABC):
    """
    Abstract node in a conversion error tree
    """

    @abc.abstractmethod
    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        """
        Print a description of this error to `file`.

        Parameters:
          indent: String to indent all extra lines we print
          inside_sum: Whether we are printing inside a [`SumErrorNode`][pane.errors.SumErrorNode] 
                      and so should omit printing the actual value we got)
          file: File-like object to print text to
        """
        ...

    def __str__(self) -> str:
        buf = StringIO()
        self.print_error(file=buf)
        return buf.getvalue().rstrip('\n')


@dataclasses.dataclass
class WrongTypeError(ErrorNode):
    expected: str
    """Short description of expected value type"""
    actual: t.Any
    """Actual value received"""
    cause: t.Optional[traceback.TracebackException] = None
    """If this was caused by an error, contains a traceback to that error"""
    info: t.Optional[str] = None
    """Additional information to supply on an new line"""

    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        if inside_sum:
            print(f"{self.expected}", file=file)
        else:
            print(f"Expected {self.expected}, instead got `{self.actual}` of type `{type(self.actual).__name__}`", file=file)
        if self.info is not None:
            print(f"{indent}{self.info}", file=file)
        if self.cause is not None:
            s = f"{indent}\n".join(self.cause.format())
            print(f"Caused by exception:\n{indent}{s}", file=file)

    def _get_cause(self) -> str:
        """Format `cause` as a human-readable string"""
        if self.cause is None:
            return 'None'
        if isinstance(self.cause, traceback.TracebackException):
            return "\n".join(self.cause.format_exception_only())
        return "\n".join(traceback.format_exception(type(self.cause), self.cause, None))

    def __repr__(self) -> str:
        return f"WrongTypeError(expected={self.expected!r}, actual={self.actual!r}, cause={self._get_cause()!r}, info={self.info!r})"

    def __eq__(self, other: t.Any) -> bool:
        # mostly useful for testing
        if not self.__class__ == other.__class__:
            return False

        return (
            self.expected == other.expected and
            self.actual == other.actual and
            self.info == other.info and
            self._get_cause() == other._get_cause()
        )


@dataclasses.dataclass
class WrongLenError(ErrorNode):
    expected: str
    """Short description of expected value type"""
    expected_len: t.Tuple[int, int]
    """(min, max) expected value length"""
    actual: t.Any
    """Actual value received"""
    actual_len: int
    """Actual length received"""

    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        len_range = '-'.join(map(str, self.expected_len))
        if inside_sum:
            print(f"{self.expected} (length {len_range})", file=file)
        else:
            print(f"Expected {self.expected} of length {len_range}, instead got `{self.actual}` of length {self.actual_len}", file=file)


@dataclasses.dataclass
class ConditionFailedError(ErrorNode):
    expected: str
    """Short description of expected value type"""
    actual: t.Any
    """Actual value received"""
    condition: str
    """Name of condition which failed"""
    cause: t.Optional[traceback.TracebackException] = None
    """If this was caused by an error, contains a traceback to that error"""

    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        if inside_sum:
            print(self.expected, end="", file=file)
        else:
            print(f"Expected {self.expected}, instead got `{self.actual}`", end="", file=file)
        if self.cause is not None:
            s = f"{indent}\n".join(self.cause.format())
            print(f"\nFailed to call condition '{self.condition}':\n{indent}{s}", file=file)
        else:
            print(f" (failed condition '{self.condition}')", file=file)


@dataclasses.dataclass
class DuplicateKeyError(ErrorNode):
    key: str
    """Offending key"""
    aliases: t.Sequence[str]
    """List of keys semantically identical to `key`"""

    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        assert not inside_sum
        print(f"Duplicate key {self.key} (same as {'/'.join(self.aliases)})", file=file)


@dataclasses.dataclass
class ProductErrorNode(ErrorNode):
    expected: str
    """Short description of expected value type"""
    children: t.Dict[t.Union[int, str], ErrorNode]
    """Map containing errors parsing subfields, if any"""
    actual: t.Any
    """Actual value received"""
    missing: t.AbstractSet[t.Union[t.Sequence[str], str]] = dataclasses.field(default_factory=set)
    """List of missing fields/equivalent aliases to fields"""
    extra: t.AbstractSet[str] = dataclasses.field(default_factory=set)
    """List of extra, unexpected fields"""

    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        # fuse together non-branching productnodes
        while len(self.children) == 1 and not len(self.missing) and not len(self.extra):
            field, child = next(iter(self.children.items()))
            if not isinstance(child, ProductErrorNode):
                break
            children: t.Dict[t.Union[str, int], ErrorNode] = {f"{field}.{k}": v for (k, v) in child.children.items()}
            missing = set(f"{field}.{f}" for f in child.missing)
            extra = set(f"{field}.{f}" for f in child.extra)
            self = ProductErrorNode(self.expected, children, self.actual, missing, extra)

        print(f"{'' if inside_sum else 'Expected '}{self.expected}", file=file)
        for (field, child) in self.children.items():
            print(f"{indent}While parsing field '{field}':\n{indent}  ", end="", file=file)
            child.print_error(f"{indent}  ", file=file)

        for field in self.missing:
            if not isinstance(field, str):
                field = '/'.join(field)
            print(f"{indent}  Missing required field '{field}'", file=file)

        for field in self.extra:
            print(f"{indent}  Unexpected field '{field}'", file=file)


@dataclasses.dataclass
class SumErrorNode(ErrorNode):
    children: t.List[ErrorNode]
    """Map containing the errors while parsing as each variant"""

    def print_error(self, indent: str = "", inside_sum: bool = False, file: t.TextIO = sys.stdout):
        def _flatten_sum(children: t.Iterable[ErrorNode]) -> t.Iterator[ErrorNode]:
            for child in children:
                if isinstance(child, SumErrorNode):
                    yield from child.children
                else:
                    yield child

        print(f"Expected one of:", file=file)
        actual = None
        for child in _flatten_sum(self.children):
            print(f"{indent}- ", end="", file=file)
            child.print_error(f"{indent}  ", inside_sum=True, file=file)
            actual = getattr(child, 'actual', actual)
        print(f"{indent}Instead got `{actual}` of type `{type(actual).__name__}`", file=file)


__all__ = [
    'ErrorNode', 'ProductErrorNode', 'SumErrorNode',
    'DuplicateKeyError', 'WrongTypeError', 'ConvertError',
    'ParseInterrupt',
]



from __future__ import annotations

import dataclasses
import itertools
import re
import typing as t

from typing_extensions import ParamSpec, Self

from .util import replace_typevars, KW_ONLY


class _Missing:
    pass


T = t.TypeVar('T')
_MISSING = _Missing()


RenameStyle = t.Literal['snake', 'camel', 'pascal', 'kebab', 'scream']
"""List of supported field-renaming styles"""


_CONVERT_FNS: t.Dict[RenameStyle, t.Callable[[t.Sequence[str]], str]] = {
    'snake': lambda parts: '_'.join(part.lower() for part in parts),
    'scream': lambda parts: '_'.join(part.upper() for part in parts),
    'kebab': lambda parts: '-'.join(part.lower() for part in parts),
    'camel': lambda parts: ''.join(part.lower() if i == 0 else part.title() for (i, part) in enumerate(parts)),
    'pascal': lambda parts: ''.join(part.title() for part in parts),
}


def _pairwise(it: t.Iterable[T]) -> t.Iterator[t.Tuple[T, T]]:
    it = iter(it)
    while True:
        try:
            a = next(it)
            b = next(it)
        except StopIteration:
            break
        yield (a, b)


def _split_field_name(field: str) -> t.Sequence[str]:
    """Split `field` into parts for renaming"""
    parts = re.split(r'[_-]', field)
    if not all(parts):
        raise ValueError(f"Unable to interpret field '{field}' for automatic rename")

    def split_case(field: str):
        if field.isupper() or field.islower() or field.istitle():
            yield field
            return
        seps = re.split(r'([A-Z])', field)
        if seps[0] != '':
            yield seps[0]
        for (s1, s2) in _pairwise(seps[1:]):
            yield s1 + s2

    return tuple(itertools.chain.from_iterable(map(split_case, parts)))


def rename_field(field: str, style: RenameStyle) -> str:
    """
    Rename `field` to match style `style`.

    Parameters:
        field: Field name to rename
        style: Style to match
    """
    return _CONVERT_FNS[style](_split_field_name(field))


@dataclasses.dataclass
class Field:
    """
    Represents a materialized dataclass field.

    Typically instantiated from a [`FieldSpec`][pane.field.FieldSpec].
    """

    _: KW_ONLY = dataclasses.field(init=False, repr=False, compare=False)
    name: str
    """Name of field"""
    type: type
    """Type of field. Must be [`Convertible`][pane.convert.Convertible]."""
    in_names: t.Sequence[str]
    """List of names which convert to this field."""
    out_name: str
    """Name this field converts into."""
    init: bool = True
    """Whether to add this field to __init__ methods (and conversion)"""
    default: t.Union[t.Any, _Missing] = _MISSING
    """Default value for field"""
    default_factory: t.Optional[t.Callable[[], t.Any]] = None
    """Default value factory for field"""
    kw_only: bool = False
    """Whether field is keyword only"""

    @classmethod
    def make(cls, name: str, ty: type,
             in_rename: t.Optional[t.Sequence[RenameStyle]] = None,
             out_rename: t.Optional[RenameStyle] = None) -> Field:
        in_names = tuple(rename_field(name, style) for style in in_rename) if in_rename is not None else (name,)
        out_name = rename_field(name, out_rename) if out_rename is not None else name
        return cls(name=name, type=ty, in_names=in_names, out_name=out_name)

    def has_default(self) -> bool:
        """Return whether this field has a default value"""
        return self.default is not _MISSING or self.default_factory is not None


@dataclasses.dataclass
class FieldSpec:
    """
    Represents a field specification.

    This hasn't been applied to a class yet, so some information is missing.

    In most cases, end users should use the [`field()`][pane.field.field] function instead.
    """

    _: KW_ONLY = dataclasses.field(init=False, repr=False, compare=False)
    rename: t.Optional[str] = None
    """Rename this field. Affects both `in_names` and `out_name`."""
    in_names: t.Optional[t.Sequence[str]] = None
    """Complete list of names which convert to this field."""
    aliases: t.Optional[t.Sequence[str]] = None
    """Additional list of names which convert to this field (excluding the name in Python)."""
    out_name: t.Optional[str] = None
    """Name this field converts into."""
    init: bool = True
    """Whether to add this field to __init__ methods (and conversion)"""
    default: t.Union[t.Any, _Missing] = _MISSING
    """Default value for field"""
    default_factory: t.Optional[t.Callable[[], t.Any]] = None
    """Default value factory for field"""
    kw_only: bool = False
    """Whether field is keyword only"""
    ty: t.Union[type, _Missing] = _MISSING
    """Type of field, if known. Must be Convertible."""

    def __post_init__(self):
        if isinstance(self.aliases, str):
            self.aliases = [self.aliases]

    def replace_typevars(self, replacements: t.Mapping[t.Union[t.TypeVar, ParamSpec], t.Type[t.Any]]) -> Self:
        """
        Apply type variable replacements to `self`.
        """
        if self.ty is _MISSING:
            return dataclasses.replace(self)
        return dataclasses.replace(self, ty=replace_typevars(t.cast(type, self.ty), replacements))

    def make_field(self, name: str,
                   in_rename: t.Optional[t.Sequence[RenameStyle]] = None,
                   out_rename: t.Optional[RenameStyle] = None) -> Field:
        """
        Make a [`Field`][pane.field.Field] from this [`FieldSpec`][pane.field.FieldSpec].
        """
        # out_name
        if self.out_name is not None:
            out_name = self.out_name
        elif self.rename is not None:
            out_name = self.rename
        else:
            out_name = rename_field(name, out_rename) if out_rename is not None else name

        if sum(p is not None for p in (self.rename, self.aliases, self.in_names)) > 1:
            raise TypeError("Can only specify one of 'rename', 'aliases', and 'in_names'")

        if self.rename is not None:
            in_names = (self.rename,)
        elif self.aliases is not None:
            in_names = (name, *(alias for alias in self.aliases if alias != name))
        elif self.in_names is not None:
            in_names = self.in_names
        else:
            in_names = tuple(rename_field(name, style) for style in in_rename) if in_rename is not None else (name,)

        ty = t.cast(type, t.Any if self.ty is _MISSING else self.ty)
        return Field(name=name, type=ty, out_name=out_name, in_names=in_names,
                     init=self.init, default=self.default, default_factory=self.default_factory,
                     kw_only=self.kw_only)


# only allow one of rename, in_names, and aliases
@t.overload
def field(*,
    rename: t.Optional[str] = None,
    in_names: None = None,
    aliases: None = None,
    out_name: t.Optional[str] = None,
    init: bool = True,
    default: t.Union[T, _Missing] = _MISSING,
    default_factory: t.Optional[t.Callable[[], T]] = None,
    kw_only: bool = False,
) -> t.Any:
    ...

@t.overload
def field(*,
    rename: None = None,
    in_names: t.Sequence[str],
    aliases: None = None,
    out_name: t.Optional[str] = None,
    init: bool = True,
    default: t.Union[T, _Missing] = _MISSING,
    default_factory: t.Optional[t.Callable[[], T]] = None,
    kw_only: bool = False,
) -> t.Any:
    ...

@t.overload
def field(*,
    rename: None = None,
    in_names: None = None,
    aliases: t.Sequence[str],
    out_name: t.Optional[str] = None,
    init: bool = True,
    default: t.Union[T, _Missing] = _MISSING,
    default_factory: t.Optional[t.Callable[[], T]] = None,
    kw_only: bool = False,
) -> t.Any:
    ...

@t.overload
def field(*,
    rename: t.Optional[str] = None,
    in_names: t.Optional[t.Sequence[str]] = None,
    aliases: t.Optional[t.Sequence[str]] = None,
    out_name: t.Optional[str] = None,
    init: bool = True,
    default: t.Union[T, _Missing] = _MISSING,
    default_factory: t.Optional[t.Callable[[], T]] = None,
    kw_only: bool = False,
) -> t.Any:
    ...

def field(*,
    rename: t.Optional[str] = None,
    in_names: t.Optional[t.Sequence[str]] = None,
    aliases: t.Optional[t.Sequence[str]] = None,
    out_name: t.Optional[str] = None,
    init: bool = True,
    default: t.Union[T, _Missing] = _MISSING,
    default_factory: t.Optional[t.Callable[[], T]] = None,
    kw_only: bool = False,
) -> t.Any:
    """
    Annotate a dataclass field.

    Parameters:
      rename: Name to rename this field as. Used for both input and output. Useful when a field name should be different inside vs. outside of Python.
      in_names: List of names which should convert into this field. If specified, the field name inside Python will be excluded (unlike `aliases`).
      aliases: List of aliases (additional names) for this field. Includes the field name inside Python (unlike `in_names`).
      out_name: Name which this field should convert into.
      init: If `False`, this field won't be touched by `pane`, and it's up to the class to initialize it in `__post_init__`.
      default: Default value for field
      default_factory: Default value factory for field
      kw_only: Whether the field is keyword-only.
    """
    return FieldSpec(
        rename=rename, in_names=in_names, aliases=aliases, out_name=out_name,
        init=init, default=default, default_factory=default_factory, kw_only=kw_only
    )


__all__ = [
    'Field', 'FieldSpec', 'field',
    'rename_field',
]



from __future__ import annotations

import re
import datetime
import collections
from decimal import Decimal
from fractions import Fraction
import pathlib
import os
import typing as t

import pytest

from pane.errors import ErrorNode, SumErrorNode, ProductErrorNode, WrongTypeError, ConditionFailedError
from pane.convert import convert, from_data, make_converter, ConvertError
from pane.converters import Converter, ScalarConverter, TupleConverter, SequenceConverter, TaggedUnionConverter, AnyConverter
from pane.converters import StructConverter, UnionConverter, LiteralConverter, ConditionalConverter, NestedSequenceConverter
from pane.converters import PatternConverter, DatetimeConverter
from pane.annotations import Condition, Tagged, val_range, len_range


class TestConvertible():
    @classmethod
    def _converter(cls, *args) -> Converter[TestConvertible]:
        return TestConverter()  # type: ignore

    def __hash__(self):
        return hash(self.__class__)

    def __eq__(self, other):
        return self.__class__ == other.__class__


class TestConverter(Converter[TestConvertible]):
    def try_convert(self, val: t.Any):
        return TestConvertible()

    def expected(self, plural: bool = False) -> str:
        return "TestConvertible"

    def collect_errors(self, val: t.Any) -> None:
        return None

    def __hash__(self):
        return hash(self.__class__)

    def __eq__(self, other):
        return self.__class__ == other.__class__


@pytest.mark.parametrize(('input', 'conv'), [
    (int, ScalarConverter(int, int, 'an int', 'ints')),
    ({'x': int, 'y': float}, StructConverter(dict, {'x': int, 'y': float})),
    (t.Tuple[int, ...], SequenceConverter(tuple, int)),
    (list[str], SequenceConverter(list, str)),
    (tuple, SequenceConverter(tuple, t.Any)),
    (t.Tuple[int, str], TupleConverter(tuple, (int, str))),
    (t.Tuple[()], TupleConverter(tuple, ())),
    (TestConvertible, TestConverter()),
    (t.Union[str, int], UnionConverter((str, int))),
    (t.Literal['a', 'b', 'c'], LiteralConverter(('a', 'b', 'c'))),
    (t.Literal['a'], LiteralConverter(('a',))),
    (re.Pattern[bytes], PatternConverter(bytes)),
    (t.Pattern, PatternConverter(str)),
    (datetime.datetime, DatetimeConverter(datetime.datetime)),
    (datetime.time, DatetimeConverter(datetime.time)),
])
def test_make_converter(input, conv: Converter):
    assert make_converter(input) == conv


@pytest.mark.parametrize(('input', 'error'), [
    (t.Pattern[int], TypeError("Pattern only accepts a 'str' or 'bytes' type argument, instead got '<class 'int'>'")),
    # tag not found, tag matches multiple types
])
def test_make_converter_raises(input, error: Exception):
    with pytest.raises(type(error)) as exc_info:
        make_converter(input)
    assert str(exc_info.value) == str(error)


@pytest.mark.parametrize(('input', 'conv', 's'), [
    (t.TypeVar('T'), AnyConverter(), "Unbound TypeVar '~T'. Will be interpreted as 'typing.Any'."),
    (t.TypeVar('U', int, str), UnionConverter((int, str)), "Unbound TypeVar '~U'. Will be interpreted as 'typing.Union[int, str]'."),
    (t.TypeVar('V', bound=tuple), SequenceConverter(tuple, t.Any), "Unbound TypeVar '~V'. Will be interpreted as '<class 'tuple'>'."),
])
def test_typevar_converters(input, conv, s):
    with pytest.warns(UserWarning, match=re.escape(s)):
        assert make_converter(input) == conv


cond1 = Condition(lambda v: True, 'true', lambda exp, plural: exp)
cond2 = Condition(lambda v: v > 0, 'v > 0', lambda exp, plural: f"{exp} > 0")


class Variant1(int):
    tag: int = 3

class Variant2(float):
    tag: int = 4

class Variant3(dict):
    tag: int = 3

class Variant4(dict):
    tag: int = 4


def test_make_converter_annotated():
    inner = int
    inner_conv = ScalarConverter(int, int, 'an int', 'ints')

    conv = make_converter(t.Annotated[int, cond1])
    assert conv == ConditionalConverter(
        int, cond1.f, 'true', cond1.make_expected
    )
    assert conv.inner == inner_conv

    conv = make_converter(t.Annotated[int, cond2])
    assert conv == ConditionalConverter(
        int, cond2.f, 'v > 0', cond2.make_expected
    )
    assert conv.inner == inner_conv

    compound = make_converter(t.Annotated[int, cond1, cond2])
    assert isinstance(compound, ConditionalConverter)
    assert compound.condition_name == 'true and v > 0'

    assert make_converter(t.Annotated[t.Union[Variant1, Variant2], Tagged('tag')]) \
        == TaggedUnionConverter((Variant1, Variant2), 'tag', external=False)

    assert make_converter(t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', ('t', 'c'))]) \
        == TaggedUnionConverter((Variant1, Variant2), 'tag', external=('t', 'c'))

    with pytest.raises(TypeError, match="Tag value '3' matches multiple types"):
        make_converter(t.Annotated[t.Union[Variant1, Variant3], Tagged('tag')])

@pytest.mark.parametrize(('conv', 'plural', 'expected'), [
    (int, False, 'an int'),
    (t.Optional[str], True, 'strings or null values'),
    (t.Sequence[str], True, 'sequences of strings'),
    (t.Dict[str, int], False, 'mapping of strings => ints'),
    (t.Literal['a', 'b', 'c'], False, "'a', 'b', or 'c'"),
    (t.Annotated[int, cond2], False, "an int > 0"),
    (t.Annotated[int, cond2], True, "ints > 0"),
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag')], False, "an int or a float"),
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', external=True)], False, "a mapping '3 or 4' => an int or a float"),
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', external=('t', 'c'))], False, "a mapping 't' => 3 or 4, 'c' => an int or a float"),
    (t.Pattern[str], True, 'string regex patterns'),
    (datetime.datetime, True, 'datetimes'),
    (datetime.time, False, 'a time'),
])
def test_converter_expected(conv: Converter, plural: bool, expected: str):
    if not isinstance(conv, Converter):
        conv = make_converter(conv)
    assert conv.expected(plural) == expected


@pytest.mark.parametrize(('ty', 'val', 'result'), [
    # scalar converters
    (int, 's', WrongTypeError('an int', 's')),
    (float, 5, 5.0),
    (complex, 6, 6.0+0j),
    # bytes types
    (bytes, b'bytestring', b'bytestring'),
    (str, b'bytestring', WrongTypeError('a string', b'bytestring')),
    (bytes, 'string', WrongTypeError('a bytestring', 'string')),
    (bytearray, b'bytestring', bytearray(b'bytestring')),
    (bytes, bytearray(b'bytestring'), b'bytestring'),
    # be conservative for now
    (t.List[int], bytearray(b'bytestring'), WrongTypeError('sequence of ints', bytearray(b'bytestring'))),
    (bytearray, [98, 121, 116, 101, 115, 116, 114, 105, 110, 103], WrongTypeError('a bytearray', [98, 121, 116, 101, 115, 116, 114, 105, 110, 103])),
    # sequence converters
    (t.Tuple[int, ...], [1, 2], (1, 2)),
    (t.Sequence[int], [1, 2], (1, 2)),  # same as tuple
    (t.List[int], (1, 2), [1, 2]),
    # exotic sequences
    (t.Set[float], (1, 2, 3), {1., 2., 3.}),
    (collections.deque, (1, 2, 3), collections.deque((1, 2, 3))),
    (t.MutableSet[int], (1, 2, 3), {1, 2, 3}),
    # dict converters
    (t.Dict[str, float], {'a': 1}, {'a': 1.0}),
    (collections.abc.Mapping[str, float], {'a': 1}, {'a': 1.0}),
    (t.MutableMapping[str, str], {'a': 'b'}, {'a': 'b'}),
    (t.Counter[str], {'a': 5, 'b': 1, 'c': 2}, collections.Counter('aaaaabcc')),
    (collections.defaultdict, {'a': 'b'}, collections.defaultdict(None, {'a': 'b'})),
    # tuple converters
    (t.Tuple[str], ['int'], ('int',)),
    (t.Tuple[str], ('s1', 's2'), WrongTypeError('tuple of length 1', ('s1', 's2'))),
    # union converters (left to right)
    (t.Union[int, float, str], 5., 5.),
    (t.Union[str, float, int], 5, 5.),
    # struct converter
    ({'x': int, 'y': float}, {'x': 5, 'y': 4}, {'x': 5, 'y': 4.}),
    ({'x': t.Union[str, int], 'y': t.Tuple[t.Union[str, int], int]}, {'x': 5., 'y': (0., 's')},
     ProductErrorNode('struct', {
         'x': SumErrorNode([WrongTypeError('a string', 5.), WrongTypeError('an int', 5.)]),
         'y': ProductErrorNode('tuple of length 2', {
               0: SumErrorNode([WrongTypeError('a string', 0.), WrongTypeError('an int', 0.)]),
               1: WrongTypeError('an int', 's'),
         },  (0., 's')),
     }, {'x': 5., 'y': (0., 's')})
    ),
    # custom convertible type
    (TestConvertible, 's', TestConvertible()),
    # conditions
    (t.Annotated[int, cond2], 1, 1),
    (t.Annotated[int, cond2], 0, ConditionFailedError('an int > 0', 0., 'v > 0')),
    (t.Annotated[int, cond1, cond2], 0, ConditionFailedError('an int satisfying true and v > 0', 0., 'true and v > 0')),
    (t.Annotated[int, val_range(min=0)], 0, 0),
    (t.Annotated[int, val_range(min=0)], -1, ConditionFailedError('an int satisfying v >= 0', -1., 'v >= 0')),
    (t.Annotated[float, val_range(min=0, max=5)], 5, 5.),
    (t.Annotated[float, val_range(min=0, max=5)], 5.05, ConditionFailedError('a float satisfying v >= 0 and v <= 5', 5.05, 'v >= 0 and v <= 5')),
    (t.Annotated[t.Sequence[int], len_range(min=1)], [], ConditionFailedError('sequence of ints with at least 1 elem', [], 'at least 1 elem')),
    # tagged unions
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', True)], {3: 4}, Variant1(4)),
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', ('t', 'c'))], {'t': 4, 'c': 4.}, Variant2(4.)),
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', ('t', 'c'))], {'t': 5, 'c': 4.}, WrongTypeError("tag 'tag' one of 3 or 4", 5)),
    (t.Annotated[t.Union[Variant1, Variant2], Tagged('tag', ('t', 'c'))], {'c': 4.}, WrongTypeError("mapping with keys 't' and 'c'", {'c': 4.})),
    (t.Annotated[t.Union[Variant3, Variant4], Tagged('tag')], {'tag': 3, 'val1': 4, 'val2': 3.}, Variant3({'val1': 4, 'val2': 3.})),
    (t.Annotated[t.Union[Variant3, Variant4], Tagged('tag')], {'tag': 5, 'v': 4}, WrongTypeError("tag 'tag' one of 3 or 4", 5)),
    # regex patterns
    (re.Pattern, 'abcde', re.compile('abcde')),
    (re.Pattern[bytes], 'abcde', WrongTypeError('a bytes regex pattern', 'abcde')),
    (re.Pattern[bytes], b'abcde', re.compile(b'abcde')),
    (re.Pattern[str], '(', WrongTypeError('a string regex pattern', '(', cause=re.error('missing ), unterminated subpattern at position 0'))),
    (re.Pattern[bytes], re.compile(b'abcde'), re.compile(b'abcde')),
    # datetime (from str)
    (datetime.datetime, "2023-09-05 11:11:11", datetime.datetime(2023, 9, 5, 11, 11, 11)),
    (datetime.time, "11:11:11", datetime.time(11, 11, 11)),
    (datetime.date, "2023-09-05", datetime.date(2023, 9, 5)),
    (datetime.date, "11:11:11", WrongTypeError('a date', "11:11:11", cause=ValueError("Invalid isoformat string: '11:11:11'"))),
    (datetime.date, "09/05/2023", WrongTypeError('a date', "09/05/2023", cause=ValueError("Invalid isoformat string: '09/05/2023'"))),
    # TODO should we accept this?
    (datetime.time, "2023-09-05 11:11:11", WrongTypeError('a time', "2023-09-05 11:11:11", cause=ValueError("Invalid isoformat string: '2023-09-05 11:11:11'"))),
    # fraction/decimal
    (Decimal, 5, Decimal('5')),
    (Decimal, '5.123', Decimal('5.123')),
    (Decimal, 1.1, Decimal('1.100000000000000088817841970012523233890533447265625')),
    (Fraction, '1/5', Fraction(1, 5)),
    (Fraction, 5.1345, Fraction(5780933071683453, 1125899906842624)),
    (Fraction, '1/0', WrongTypeError('a fraction', '1/0', cause=ZeroDivisionError('Fraction(1, 0)'))),
    # path
    (pathlib.PurePosixPath, "/test/path", pathlib.PurePosixPath("/test/path")),
    (os.PathLike, "test/path", pathlib.PurePath("test/path")),
])
def test_convert(ty, val, result):
    if isinstance(result, ErrorNode):
        with pytest.raises(ConvertError) as exc_info:
            convert(val, ty)
        assert exc_info.value.tree == result
    else:
        assert convert(val, ty) == result
        assert make_converter(ty).collect_errors(val) is None

@pytest.mark.parametrize(('conv', 'val', 'result'), [
    # to time
    (datetime.time, datetime.time.fromisoformat("11:11:11"), datetime.time.fromisoformat("11:11:11")),
    (datetime.time, datetime.date.fromisoformat("2023-09-01"), WrongTypeError('a time', datetime.date.fromisoformat("2023-09-01"))),
    (datetime.time, datetime.datetime.fromisoformat("2023-09-01 11:11:11"), datetime.time.fromisoformat("11:11:11")),
    # to date
    (datetime.date, datetime.date.fromisoformat("2023-09-01"), datetime.date.fromisoformat("2023-09-01")),
    (datetime.date, datetime.time.fromisoformat("11:11:11"), WrongTypeError('a date', datetime.time.fromisoformat("11:11:11"))),
    (datetime.date, datetime.datetime.fromisoformat("2023-09-01 11:11:11"), datetime.date.fromisoformat("2023-09-01")),
    # to datetime
    (datetime.datetime, datetime.date.fromisoformat("2023-09-01"), datetime.datetime.fromisoformat("2023-09-01 00:00:00")),
    (datetime.datetime, datetime.time.fromisoformat("12:05:23"), WrongTypeError('a datetime', datetime.time.fromisoformat("12:05:23"))),
    (datetime.datetime, datetime.datetime.fromisoformat("2023-09-01 11:11:11"), datetime.datetime.fromisoformat("2023-09-01 11:11:11")),
])
def test_conv_convert(conv, val, result):
    # this test is useful for testing how Converters respond to
    # values other than data interchange values

    if not isinstance(conv, Converter):
        conv = make_converter(conv)

    if isinstance(result, ErrorNode):
        with pytest.raises(ConvertError) as exc_info:
            conv.convert(val)
        assert exc_info.value.tree == result
    else:
        assert conv.convert(val) == result
        assert conv.collect_errors(val) is None

# TODO test converter idempotence

@pytest.mark.parametrize(('err', 's'), [
    (WrongTypeError('an int', 3.), "Expected an int, instead got `3.0` of type `float`"),
    # TODO more
])
def test_error_print(err, s):
    assert str(err) == s



import typing as t

# pyright: reportUnknownMemberType=none

try:
    from numpy import generic, dtype, ndarray, array
    from numpy.typing import NDArray
    import numpy as _numpy

    if t.TYPE_CHECKING:
        from ..converters import Converter


    def _dtype_map(ty: t.Union[t.Type[t.Any], t.Type[generic]]) -> type:
        # TODO add a lookup table here
        # TODO add conditions to some types
        # e.g. unsigned int -> NonNegativeInt
        if ty in (_numpy.generic, _numpy.object_, t.Any):
            return t.Any
        if issubclass(ty, (_numpy.integer, int)):
            return int
        if issubclass(ty, (_numpy.floating, float)):
            return float
        if issubclass(ty, (_numpy.complexfloating, complex)):
            return complex
        if issubclass(ty, (_numpy.bool_, bool)):
            return bool
        if issubclass(ty, (_numpy.bytes_, bytes)):
            return bytes
        if issubclass(ty, (_numpy.str_, str)):
            return str
        raise TypeError(f"Don't know how to handle numpy dtype '{ty}'")


    def numpy_converter_handler(ty: t.Any, args: t.Sequence[t.Any]) -> 'Converter[t.Any]':
        from ..convert import make_converter

        if issubclass(ty, generic):
            # dtype converters
            return make_converter(_dtype_map(ty))

        if not (issubclass(ty, ndarray) or ty is NDArray):
            return NotImplemented

        if issubclass(ty, ndarray):
            arg1 = t.Any if len(args) < 1 else args[0]
            dtype = t.Any if len(args) < 2 else args[1]

            if arg1 is not t.Any:
                raise TypeError("Numpy shape types are currently unsupported.")

            if dtype is not t.Any:
                dtype_ty, dtype_args = t.get_origin(dtype), t.get_args(dtype)
                if dtype_ty is not _numpy.dtype:
                    raise TypeError(f"ndarray type argument should be 'numpy.dtype[<type>]', not '{dtype}'")
                dtype = t.Any if len(dtype_args) < 1 else dtype_args[0]
        else:
            dtype = t.Any if len(args) < 1 else args[0]

        from ..converters import NestedSequenceConverter

        return NestedSequenceConverter(dtype, array, ragged=False)

except ImportError:
    if not t.TYPE_CHECKING:
        class generic():
            pass

        _DTypeScalar_co = t.TypeVar("_DTypeScalar_co", covariant=True, bound=generic)

        class dtype(t.Generic[_DTypeScalar_co]):
            ...

        _ShapeType = t.TypeVar('_ShapeType', bound=dtype[t.Any])
        _DType_co = t.TypeVar('_DType_co', bound=t.Any, covariant=True)

        class ndarray(t.Generic[_ShapeType, _DType_co]):
            pass

        ScalarType = t.TypeVar("ScalarType", covariant=True, bound=generic)
        NDArray = ndarray[t.Any, dtype[ScalarType]]

        # dummy handler
        def numpy_converter_handler(ty: t.Any, args: t.Sequence[t.Any]):
            return NotImplemented


from __future__ import annotations

import typing as t

import pytest

from pane.convert import convert, make_converter, ConvertError
from pane.types import PositiveInt, ListNotEmpty, ValueOrList, Range
from pane.errors import ConvertError, ProductErrorNode, SumErrorNode, WrongTypeError, WrongLenError, ConditionFailedError, ErrorNode


@pytest.mark.parametrize(('ty', 'val', 'result'), [
    (PositiveInt, 5, 5),
    (PositiveInt, -5, ConditionFailedError('a positive int', -5, 'positive')),
    (ListNotEmpty[int], [], ConditionFailedError('sequence of ints with at least 1 elem', [], 'at least 1 elem')),
    (Range[float], (2., 4., 2), Range[t.Any](2., 4., 2)),
    (Range[float], (2., 4., 2, 5, 8), WrongLenError('tuple Range', (2, 3), (2., 4., 2, 5, 8), 5)),
    (Range[float], {'start': 2., 'end': 4., 'step': 1.}, Range[t.Any](start=2., end=4., step=1.)),
    (Range[float], {'start': 2., 'end': 4.}, WrongTypeError('struct Range', {'start': 2., 'end': 4.}, cause=TypeError("Either 'n' or 'step' must be specified"))),
    (Range[float], {'start': 2., 'end': 4., 'n': 1, 'step': 3.}, WrongTypeError('struct Range', {'start': 2., 'end': 4., 'n': 1, 'step': 3.}, cause=TypeError("Either 'n' or 'step' may be specified, but not both"))),
    (ValueOrList[int], 5, ValueOrList.from_val(5)),
    (ValueOrList[int], (5, 6, 7), ValueOrList.from_list([5, 6, 7])),
    (ValueOrList[int], (5, 6.5, 7), SumErrorNode([WrongTypeError('an int', (5, 6.5, 7)), ProductErrorNode('sequence of ints', {1: WrongTypeError('an int', 6.5)}, (5, 6.5, 7))])),
    (ValueOrList[int], {}, SumErrorNode([WrongTypeError('an int', {}), WrongTypeError('sequence of ints', {})])),
])
def test_convert_types(ty, val, result):
    if isinstance(result, ErrorNode):
        with pytest.raises(ConvertError) as exc_info:
            convert(val, ty)
        assert exc_info.value.tree == result
    else:
        assert convert(val, ty) == result
        assert make_converter(ty).collect_errors(val) is None



"""
Pane dataclasses.
"""

from __future__ import annotations

import dataclasses
from dataclasses import FrozenInstanceError
import functools
from inspect import Signature, Parameter
import traceback
import typing as t

from typing_extensions import dataclass_transform, ParamSpec, Self

from .convert import DataType, Convertible, from_data, into_data, convert
from .converters import Converter, make_converter
from .errors import ConvertError, ParseInterrupt, ErrorNode
from .errors import WrongTypeError, WrongLenError, ProductErrorNode, DuplicateKeyError
from .field import Field, FieldSpec, field, RenameStyle, _MISSING
from .util import FileOrPath, open_file, get_type_hints, list_phrase, KW_ONLY


T = t.TypeVar('T')
PaneBaseT = t.TypeVar('PaneBaseT', bound='PaneBase')


@dataclass_transform(
    eq_default=True,
    order_default=True,
    frozen_default=True,
    kw_only_default=False,
    field_specifiers=(FieldSpec, field),
)
class PaneBase:
    """
    Base class for all `pane` dataclasses
    """
    __slots__ = ('__pane_set__',)

    __pane_info__: PaneInfo
    """Dunder attribute holding [`PaneInfo`][pane.classes.PaneInfo]"""
    __pane_set__: t.Set[str]
    """Dunder attribute holding a set of fields which have been set/modified"""

    def __init_subclass__(
        cls,
        *args: t.Any,
        name: t.Optional[str] = None,
        out_format: t.Optional[ClassLayout] = None,
        in_format: t.Optional[t.Sequence[ClassLayout]] = None,
        eq: t.Optional[bool] = None,
        order: t.Optional[bool] = None,
        frozen: t.Optional[bool] = None,
        init: t.Optional[bool] = None,
        kw_only: t.Optional[bool] = None,
        rename: t.Optional[RenameStyle] = None,
        in_rename: t.Optional[t.Union[RenameStyle, t.Sequence[RenameStyle]]] = None,
        out_rename: t.Optional[RenameStyle] = None,
        allow_extra: t.Optional[bool] = None,
        **kwargs: t.Any,
    ):
        old_params = getattr(cls, '__parameters__', ())
        super().__init_subclass__(*args, **kwargs)
        setattr(cls, '__parameters__', old_params + getattr(cls, '__parameters__', ()))

        if rename is not None:
            if in_rename is not None or out_rename is not None:
                raise ValueError("'rename' cannot be specified with 'in_rename' or 'out_rename'")
            in_rename = t.cast(t.Tuple[RenameStyle, ...], (rename,))
            out_rename = rename
        elif in_rename is not None and isinstance(in_rename, str):
            in_rename = t.cast(t.Tuple[RenameStyle, ...], (in_rename,))

        # handle option inheritance
        opts: PaneOptions = getattr(cls, PANE_INFO).opts if hasattr(cls, PANE_INFO) else PaneOptions()
        opts = opts.replace(
            name=name, out_format=out_format, in_format=in_format,
            eq=eq, order=order, frozen=frozen, init=init, allow_extra=allow_extra,
            kw_only=kw_only, in_rename=in_rename, out_rename=out_rename,
        )

        _process(cls, opts)

    def __class_getitem__(cls, params: t.Union[type, t.Tuple[type, ...]]):
        if not isinstance(params, tuple):
            params = (params,)
        return _make_subclass(cls, params)

    def __repr__(self) -> str:
        inside = ", ".join(f"{field.name}={getattr(self, field.name)!r}" for field in self.__pane_info__.fields)
        return f"{self.__class__.__name__}({inside})"

    def __setattr__(self, name: str, value: t.Any) -> None:
        opts = self.__pane_info__.opts
        if opts.frozen:
            raise FrozenInstanceError(f"cannot assign to field {name!r}")
        super().__setattr__(name, value)
        set_fields: t.Set[str] = getattr(self, PANE_SET_FIELDS)
        set_fields.add(name)

    def __delattr__(self, name: str) -> None:
        raise AttributeError(f"cannot delete field {name!r}")

    @classmethod
    def _converter(cls: t.Type[PaneBaseT], *args: t.Type[Convertible]) -> Converter[PaneBaseT]:
        if len(args) > 0:
            cls = t.cast(t.Type[PaneBaseT], cls[tuple(args)])  # type: ignore
        return PaneConverter(cls)

    @classmethod
    def from_obj(cls, obj: Convertible) -> Self:
        """
        Convert `obj` into `cls`. Equivalent to `convert(obj, cls)`

        Parameters:
          obj: Object to convert. Must be convertible.
        """
        return convert(obj, cls)

    @classmethod
    def from_data(cls, data: DataType) -> Self:
        """
        Convert `data` into `cls`. Equivalent to `from_data(data, cls)`

        Parameters:
          data: Data to convert. Must be a data interchange type.
        """
        return from_data(data, cls)

    def into_data(self) -> DataType:
        """Convert `self` into interchange data"""
        return into_data(self, self.__class__)

    def dict(self, set_only: bool = False) -> t.Dict[str, t.Any]:
        """
        Return a dict of the fields in `self`

        Parameters:
          set_only: If `True`, return only the fields which have been set
        """
        if set_only:
            return {
                k : getattr(self, k) for k in getattr(self, PANE_SET_FIELDS)
            }
        return {
            field.name: getattr(self, field.name) for field in self.__pane_info__.fields
        }

    @classmethod
    def from_json(cls, f: FileOrPath) -> Self:
        """
        Load `cls` from a JSON file `f`

        Parameters:
          f: File-like or path-like to load from
        """
        import json
        with open_file(f) as f:
            obj = json.load(f)
        return cls.from_data(obj)

    @classmethod
    def from_jsons(cls, s: str) -> Self:
        """
        Load `cls` from a JSON string `s`

        Parameters:
          s: JSON string to load from
        """
        import json
        obj = json.loads(s)
        return cls.from_data(obj)

    @classmethod
    def from_yaml(cls, f: FileOrPath) -> Self:
        """
        Load `cls` from a YAML file `f`

        Parameters:
          f: File-like or path-like to load from
        """
        import yaml
        try:
            from yaml import CSafeLoader as Loader
        except ImportError:
            from yaml import SafeLoader as Loader

        with open_file(f) as f:
            obj = list(yaml.load_all(f, Loader))

        return cls.from_data(obj)

    @classmethod
    def from_yamls(cls, s: str) -> Self:
        """
        Load `cls` from a YAML string `s`

        Parameters:
          s: YAML string to load from
        """
        from io import StringIO
        return cls.from_yaml(StringIO(s))

    @classmethod
    def make_unchecked(cls, *args: t.Any, **kwargs: t.Any) -> Self:
        ...

    @classmethod
    def from_dict_unchecked(cls, d: t.Dict[str, t.Any]) -> Self:
        ...


@dataclasses.dataclass
class PaneInfo:
    """Structure holding internal information about a `pane` dataclass"""
    opts: PaneOptions
    """Dataclass options"""
    specs: t.Dict[str, FieldSpec]
    """
    Dict of raw field specifications

    This is used by subclasses to build [`Field`][pane.field.Field]s
    """
    fields: t.Tuple[Field, ...]
    """
    Tuple of processed [`Field`][pane.field.Field]s
    """
    pos_args: t.Tuple[int, int]
    """
    Range of allowed positional argument numbers, `[min, max]` inclusive
    """


@dataclasses.dataclass(frozen=True)
class PaneOptions:
    name: t.Optional[str] = None
    """Dataclass name"""
    _: KW_ONLY = dataclasses.field(init=False, repr=False, compare=False)
    eq: bool = True
    """Whether to generate `__eq__`/`__ne__` methods"""
    order: bool = True
    """Whether to generate `__gt__`/`__ge__`/`__lt__`/`__le__` methods"""
    frozen: bool = True
    """Whether dataclass fields are frozen"""
    init: bool = True
    """Whether to generate `__init__` method"""
    kw_only: bool = False
    """Whether all fields should be keyword-only"""
    out_format: ClassLayout = 'struct'
    """Data format to convert class into"""
    in_format: t.Sequence[ClassLayout] = ('struct',)
    """Set of data formats class is convertible from"""
    in_rename: t.Optional[t.Sequence[RenameStyle]] = None
    """Set of rename styles class is convertible from"""
    out_rename: t.Optional[RenameStyle] = None
    """Rename style to convert class into"""
    allow_extra: bool = False
    """Whether extra fields are allowed in conversion"""

    def replace(self, **changes: t.Any):
        """Return `self` with the given changes applied"""
        changes['name'] = changes.get('name', None)
        return dataclasses.replace(self, **{k: v for (k, v) in changes.items() if v is not None})


@functools.lru_cache(maxsize=256)
def _make_subclass(cls: t.Any, params: t.Tuple[t.Any, ...]) -> type:
    sup: t.Any = super(PaneBase, cls)
    if not hasattr(sup, '__class_getitem__'):
        raise TypeError(f"type '{cls}' is not subscriptable")
    alias: t.Type[PaneBase] = sup.__class_getitem__(params)  # type: ignore
    typevars = getattr(cls, '__parameters__', ())

    # return subclass with bound type variables
    bound_vars = dict(zip(typevars, params))
    return type(cls.__name__, (cls,), {
        PANE_BOUNDVARS: bound_vars,
        '__origin__': cls,
        '__parameters__': getattr(alias, '__parameters__'),
    })


def _make_init(cls: t.Type[PaneBase], fields: t.Sequence[Field]):
    params: t.List[Parameter] = []
    for field in fields:
        if field.default is not _MISSING:
            default = field.default
        elif field.default_factory is not None:
            default = field.default_factory()
        else:
            default = Parameter.empty
        kind = Parameter.KEYWORD_ONLY if field.kw_only else Parameter.POSITIONAL_OR_KEYWORD
        annotation = field.type if field.type is not _MISSING else Parameter.empty
        params.append(Parameter(field.name, kind, default=default, annotation=annotation))

    sig = Signature(params, return_annotation=None)

    def __init__(self: PaneBase, *args: t.Any, **kwargs: t.Any):
        from_dict = kwargs.pop('_pane_from_dict', None)
        if from_dict is not None:
            for (k, v) in from_dict.items():
                object.__setattr__(self, k, v)
            object.__setattr__(self, PANE_SET_FIELDS, set(from_dict.keys()))
            if hasattr(self, POST_INIT):
                getattr(self, POST_INIT)()
            return

        checked = kwargs.pop('_pane_checked', True)
        try:
            bound_args = sig.bind(*args, **kwargs).arguments
        except TypeError as e:
            raise TypeError(*e.args) from None

        set_fields: t.Set[str] = set()

        for field in self.__pane_info__.fields:
            if field.name in bound_args:
                val = bound_args[field.name]
                if checked:
                    val = convert(val, field.type)  # type: ignore
                set_fields.add(field.name)
            elif field.default is not _MISSING:
                val = field.default
            elif field.default_factory is not None:
                val = field.default_factory()
            else:
                raise RuntimeError("Mismatch between fields and signature. This shouldn't happen")
            object.__setattr__(self, field.name, val)

        object.__setattr__(self, PANE_SET_FIELDS, set_fields)

        if hasattr(self, POST_INIT):
            getattr(self, POST_INIT)()

    setattr(__init__, '__signature__', sig)
    setattr(cls, '__init__', __init__)
    setattr(cls, '__signature__', sig)

    @classmethod
    def make_unchecked(cls, *args, **kwargs):  # type: ignore
        return cls(*args, **kwargs, _pane_checked=False)  # type: ignore

    sig2 = Signature([Parameter('cls', Parameter.POSITIONAL_OR_KEYWORD), *params], return_annotation=cls)
    make_unchecked.__func__.__signature__ = sig2  # type: ignore
    setattr(cls, 'make_unchecked', make_unchecked)

    @classmethod
    def from_dict_unchecked(cls, d):  # type: ignore
        return cls(_pane_from_dict=d)  # type: ignore

    sig2 = Signature([Parameter('cls', Parameter.POSITIONAL_OR_KEYWORD), Parameter('d', Parameter.POSITIONAL_OR_KEYWORD, annotation=t.Dict[str, t.Any])], return_annotation=cls)
    from_dict_unchecked.__func__.__signature__ = sig2  # type: ignore
    setattr(cls, 'from_dict_unchecked', from_dict_unchecked)


def _make_eq(cls: t.Type[PaneBase], fields: t.Sequence[Field]):
    #eq_fields = list(filter(lambda f: f.eq, fields))
    def __eq__(self: PaneBase, other: t.Any) -> bool:
        # check if classes are the same (modulo type variables)
        if self.__class__.__dict__.get('__origin__', self.__class__) != other.__class__.__dict__.get('__origin__', other.__class__):
            return False
        return all(
            getattr(self, field.name) == getattr(other, field.name)
            for field in fields
        )

    setattr(cls, '__eq__', __eq__)


def _make_ord(cls: t.Type[PaneBase], fields: t.Sequence[Field]):
    #ord_fields = list(filter(lambda f: f.ord, fields))
    def _pane_ord(self: PaneBase, other: t.Any) -> t.Literal[-1, 0, 1]:
        if self.__class__ != other.__class__:
            return NotImplemented  # type: ignore
        for field in fields:
            if getattr(self, field.name) == getattr(other, field.name):
                continue
            return 1 if getattr(self, field.name) > getattr(other, field.name) else -1
        return 0

    def __lt__(self: PaneBase, other: t.Any) -> bool:
        return NotImplemented if (o := _pane_ord(self, other)) is NotImplemented else t.cast(int, o) < 0

    def __le__(self: PaneBase, other: t.Any) -> bool:
        return NotImplemented if (o := _pane_ord(self, other)) is NotImplemented else t.cast(int, o) <= 0

    def __gt__(self: PaneBase, other: t.Any) -> bool:
        return NotImplemented if (o := _pane_ord(self, other)) is NotImplemented else t.cast(int, o) > 0

    def __ge__(self: PaneBase, other: t.Any) -> bool:
        return NotImplemented if (o := _pane_ord(self, other)) is NotImplemented else t.cast(int, o) >= 0

    setattr(cls, '_pane_ord', _pane_ord)
    setattr(cls, '__lt__', __lt__)
    setattr(cls, '__le__', __le__)
    setattr(cls, '__gt__', __gt__)
    setattr(cls, '__ge__', __ge__)


def _process(cls: t.Type[PaneBase], opts: PaneOptions):
    fields: t.List[Field] = []

    specs: t.Dict[str, FieldSpec] = {}

    # collect FieldSpecs from base classes
    for base in reversed(cls.__mro__[1:]):
        if not hasattr(base, PANE_INFO):
            continue  # not a pane dataclass
        cls_specs = getattr(base, PANE_INFO).specs

        # apply typevar replacements
        bound_vars = t.cast(t.Mapping[t.Union[t.TypeVar, ParamSpec], type], getattr(base, PANE_BOUNDVARS, {}))
        specs.update(cls_specs)
        specs = {k: spec.replace_typevars(bound_vars) for (k, spec) in specs.items()}

    annotations = get_type_hints(cls)
    kw_only = opts.kw_only  # current kw_only state
    cls_specs = {}

    for name, ty in annotations.items():
        if ty is KW_ONLY:
            # all further params are kw_only
            kw_only = True
            continue

        if isinstance(getattr(cls, name, None), FieldSpec):
            # process existing FieldSpec
            spec: FieldSpec = getattr(cls, name)
            spec.ty = ty
        else:
            # make new spec
            spec = FieldSpec(ty=ty, default=getattr(cls, name, _MISSING))

        spec.kw_only |= kw_only
        cls_specs[name] = spec

    # apply typevar replacements
    bound_vars = getattr(cls, PANE_BOUNDVARS, {})
    specs.update(cls_specs)
    specs = {k: spec.replace_typevars(bound_vars) for (k, spec) in specs.items()}

    # bake FieldSpecs into Fields
    fields = [spec.make_field(name, opts.in_rename, opts.out_rename) for (name, spec) in specs.items()]
    # reorder kw-only fields to end
    fields = [*filter(lambda f: not f.kw_only, fields), *filter(lambda f: f.kw_only, fields)]

    # positional argument lengths
    min_len, max_len = (0, 0)
    seen_opt = False
    for field in fields:
        if field.kw_only:
            if not field.has_default() and 'tuple' in opts.in_format:
                raise TypeError(f"Field '{field.name}' is kw_only but mandatory. This is incompatible with the 'tuple' in_format.")
            continue
        max_len += 1
        if field.has_default():
            seen_opt = True
        else:
            if seen_opt:
                raise TypeError(f"Mandatory field '{field.name}' follows optional field")
            min_len = max_len  # expand min length

    cls.__pane_info__ = PaneInfo(
        opts=opts, specs=cls_specs, fields=tuple(fields), pos_args=(min_len, max_len)
    )

    for field in fields:
        name = str(field.name)
        # remove Fields from class, and set defaults
        if field.default is _MISSING:
            if isinstance(getattr(cls, name, None), Field):
                delattr(cls, name)
        else:
            setattr(cls, name, field.default)

    if opts.init:
        _make_init(cls, fields)
    if opts.eq:
        _make_eq(cls, fields)
    if opts.order:
        _make_ord(cls, fields)

    return cls


class PaneConverter(Converter[PaneBaseT]):
    """
    [`Converter`][pane.converters.Converter] for `pane` dataclasses
    """
    def __init__(self, cls: t.Type[PaneBaseT]):
        self.cls = cls
        self.name = self.cls.__name__
        self.cls_info: PaneInfo = getattr(self.cls, PANE_INFO)
        self.opts: PaneOptions = self.cls_info.opts
        self.fields: t.Sequence[Field] = self.cls_info.fields
        self.field_converters: t.Sequence[Converter[t.Any]] = [make_converter(field.type) for field in self.fields]
        self.field_map: t.Dict[str, int] = {}

        for (i, field) in enumerate(self.fields):
            self.field_map[field.name] = i
            for alias in field.in_names:
                self.field_map[alias] = i

    def into_data(self, val: t.Any) -> DataType:
        """Convert dataclass `val` into data interchange, using the correct 'out_format'"""
        assert isinstance(val, PaneBase)
        if self.opts.out_format == 'tuple':
            return tuple(
                conv.into_data(getattr(val, field.name))
                for (field, conv) in zip(self.fields, self.field_converters)
            )
        elif self.opts.out_format == 'struct':
            return {
                field.out_name: conv.into_data(getattr(val, field.name))
                for (field, conv) in zip(self.fields, self.field_converters)
            }
        raise ValueError(f"Unknown 'out_format' '{self.opts.out_format}'")

    def expected(self, plural: bool = False) -> str:
        """Expected value for this converter"""
        return f"{list_phrase(self.opts.in_format)} {self.name}"

    def try_convert(self, val: t.Any) -> PaneBaseT:
        """
        See [`Converter.try_convert`][pane.converters.Converter.try_convert]

        Dispatches to [`try_convert_tuple`][pane.classes.PaneConverter.try_convert_tuple]
        and [`try_convert_struct`][pane.classes.PaneConverter.try_convert_struct]
        """
        # based on type, try to delegate to try_convert_tuple or try_convert_struct
        if isinstance(val, (list, tuple, t.Sequence)):
            val = t.cast(t.Sequence[t.Any], val)
            if 'tuple' not in self.opts.in_format:
                raise ParseInterrupt()

            return self.try_convert_tuple(t.cast(t.Sequence[t.Any], val))

        elif isinstance(val, (dict, t.Mapping)):
            if 'struct' not in self.opts.in_format:
                raise ParseInterrupt()

            return self.try_convert_struct(t.cast(t.Mapping[str, t.Any], val))

        raise ParseInterrupt()

    def collect_errors(self, val: t.Any) -> t.Union[WrongTypeError, WrongLenError, ProductErrorNode, None]:
        """
        See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]

        Dispatches to [`collect_errors_tuple`][pane.classes.PaneConverter.collect_errors_tuple]
        and [`collect_errors_struct`][pane.classes.PaneConverter.collect_errors_struct]
        """
        # based on type, try to delegate to collect_errors_tuple or collect_errors_struct
        if isinstance(val, (list, tuple, t.Sequence)):
            if 'tuple' not in self.opts.in_format:
                return WrongTypeError(self.expected_struct(), val)

            return self.collect_errors_tuple(t.cast(t.Sequence[t.Any], val))

        elif isinstance(val, (dict, t.Mapping)):
            val = t.cast(t.Mapping[str, t.Any], val)
            if 'struct' not in self.opts.in_format:
                return WrongTypeError(f'tuple {self.name}', val)
            
            return self.collect_errors_struct(t.cast(t.Mapping[str, t.Any], val))

        return WrongTypeError(self.name, val)

    def expected_struct(self, plural: bool = False) -> str:
        """Expected value for the 'struct' data format"""
        return f"struct {self.name}"

    def try_convert_struct(self, val: t.Mapping[str, t.Any]) -> PaneBaseT:
        """[`Converter.try_convert`][pane.converters.Converter.try_convert] for the 'struct' data format"""
        # loop through values, and handle accordingly
        values: t.Dict[str, t.Any] = {}
        for (k, v) in t.cast(t.Dict[str, t.Any], val).items():
            if k not in self.field_map:
                if not self.opts.allow_extra:
                    raise ParseInterrupt()  # extra key
                continue
            field = self.fields[self.field_map[k]]
            conv = self.field_converters[self.field_map[k]]

            if field.name in values:
                raise ParseInterrupt()  # multiple values for key
            values[field.name] = conv.try_convert(v)

        for field in filter(lambda field: field.name not in values, self.fields):
            if field.default is not _MISSING:
                values[field.name] = field.default
            elif field.default_factory is not None:
                values[field.name] = field.default_factory
            else:
                raise ParseInterrupt()  # missing field

        try:
            return self.cls.from_dict_unchecked(values)
        except Exception:  # error in __post_init__
            raise ParseInterrupt()

    def collect_errors_struct(self, val: t.Mapping[str, t.Any]) -> t.Union[WrongTypeError, ProductErrorNode, None]:
        """[`Converter.collect_errors`][pane.converters.Converter.collect_errors] for the 'struct' data format"""
        values: t.Dict[str, t.Any] = {}  # converted field values. Required to check for __post_init__ errors
        children: t.Dict[t.Union[str, int], ErrorNode] = {}  # errors in converting fields
        extra: t.Set[str] = set()  # extra fields found
        seen: t.Set[str] = set()   # fields seen already (used to find dupes)
        for (k, v) in val.items():
            if k not in self.field_map:
                if not self.opts.allow_extra:
                    extra.add(k)  # unknown key
                continue

            field = self.fields[self.field_map[k]]
            conv = self.field_converters[self.field_map[k]]
            if field.name in seen:
                children[k] = DuplicateKeyError(k, field.in_names)
                continue
            seen.add(field.name)

            # this is a little tricky. we need to call convert() rather
            # than collect_errors to grab a successful value
            try:
                values[field.name] = conv.convert(v)
            except ConvertError as e:
                # then we can collect errors if that fails.
                children[k] = e.tree

        missing: t.Set[str] = set()
        for field in self.fields:
            if field.name not in seen and not field.has_default():
                missing.add(field.name)

        if len(missing) or len(children) or len(extra):
            # return field errors
            return ProductErrorNode(self.expected_struct(), children, val, missing, extra)
        try:
            self.cls.make_unchecked(**values)
            return None
        except Exception as e:  # error in __post_init__
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return WrongTypeError(f'struct {self.name}', val, tb)

    def expected_tuple(self, plural: bool = False) -> str:
        """Expected value for the 'tuple' data format"""
        return f"tuple {self.name}"

    def try_convert_tuple(self, val: t.Sequence[t.Any]) -> PaneBaseT:
        """[`Converter.try_convert`][pane.converters.Converter.try_convert] for the 'tuple' data format"""
        (min_len, max_len) = self.cls_info.pos_args
        if min_len < len(val) > max_len:
            raise ParseInterrupt()

        vals: t.List[t.Any] = []
        for (conv, v) in zip(self.field_converters, val):
            vals.append(conv.try_convert(v))

        try:
            return self.cls.make_unchecked(*vals)
        except Exception:  # error in __post_init__
            raise ParseInterrupt()

    def collect_errors_tuple(self, val: t.Sequence[t.Any]) -> t.Union[WrongTypeError, ProductErrorNode, WrongLenError, None]:
        """[`Converter.collect_errors`][pane.converters.Converter.collect_errors] for the 'tuple' data format"""
        (min_len, max_len) = self.cls_info.pos_args
        if min_len < len(val) > max_len:
            return WrongLenError(f'tuple {self.name}', (min_len, max_len), val, len(val))

        vals: t.List[t.Any] = []
        children: t.Dict[t.Union[str, int], ErrorNode] = {}
        for (i, (conv, v)) in enumerate(zip(self.field_converters, val)):
            # this is a little tricky. we need to call convert() rather
            # than collect_errors to grab a successful value
            try:
                vals.append(conv.convert(v))
            except ConvertError as e:
                # then we can collect errors if that fails.
                children[i] = e.tree

        if len(children):
            # return field errors
            return ProductErrorNode(self.expected_tuple(), children, val)

        try:
            self.cls.make_unchecked(*vals)
            return None
        except Exception as e:  # error in __post_init__
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return WrongTypeError(f'tuple {self.name}', val, tb)


ClassLayout = t.Literal['tuple', 'struct']
"""Set of known class layouts for 'in_formats' and 'out_format'."""
PANE_INFO = '__pane_info__'  # class information
"""Name of dunder attribute holding [`PaneInfo`][pane.classes.PaneInfo]"""
PANE_SET_FIELDS = '__pane_set__'  # fields currently set
"""Name of dunder attribute holding a set of fields which have been set/modified"""
PANE_BOUNDVARS = '__pane_boundvars__'  # bound variables, for subtypes of generics
"""Name of dunder attribute holding a dictionary of bound type variables (for generic subclasses only)."""
POST_INIT = '__post_init__'  # post-init function
"""Name of post-init method"""


__all__ = [
    'PaneBase', 'PaneOptions', 'field', 'Field', 'KW_ONLY',
]



"""Helper types for use with ``pane.convert`` and dataclasses."""
from __future__ import annotations

import math
import typing as t

from pane.classes import PaneBase, field
from pane.converters import UnionConverter
from pane.convert import Convertible, DataType, into_data
from pane.annotations import *


T = t.TypeVar('T', bound=Convertible)
U = t.TypeVar('U', bound=Convertible)
Num = t.TypeVar('Num', bound=t.Union[int, float])

PositiveInt = t.Annotated[int, Positive]
NonNegativeInt = t.Annotated[int, NonNegative]
NegativeInt = t.Annotated[int, Negative]
NonPositiveInt = t.Annotated[int, NonPositive]

PositiveFloat = t.Annotated[float, Positive]
NonNegativeFloat = t.Annotated[float, NonNegative]
NegativeFloat = t.Annotated[float, Negative]
NonPositiveFloat = t.Annotated[float, NonPositive]
FiniteFloat = t.Annotated[float, Finite]

ListNotEmpty = t.Annotated[t.List[T], len_range(min=1)]


class Range(PaneBase, t.Generic[Num],
            in_format=('tuple', 'struct'),
            out_format='struct'):
    start: Num
    end: Num

    n: t.Optional[NonNegativeInt] = field(default=None)
    step: t.Optional[Num] = field(default=None, kw_only=True)

    def __post_init__(self):
        s = sum((self.step is None, self.n is None))
        if s == 0:
            raise TypeError("Either 'n' or 'step' may be specified, but not both")
        if s == 2:
            raise TypeError("Either 'n' or 'step' must be specified")
        span = self.end - self.start
        if self.step is not None:
            if math.isclose(self.step, 0.):
                raise ValueError("'step' should be nonzero")
            n = 1 + math.ceil(span / self.step - 1e-6) if span > 0 else 0
            object.__setattr__(self, 'n', n)
        else:
            assert self.n is not None
            if not isinstance(self.start, float) and span % (self.n - 1):
                raise ValueError("Range must be evenly divisible by 'n'")
            step = type(self.start)(span / (self.n - 1)) if self.n > 1 else None
            object.__setattr__(self, 'step', step)

    def __len__(self) -> int:
        return t.cast(int, self.n)

    def __iter__(self) -> t.Iterator[Num]:
        assert self.n is not None
        if self.n == 0:
            return
        val: Num = self.start
        for _ in range(self.n - 1):
            yield val
            val = t.cast(Num, val + self.step)
        yield self.end


class ValueOrList(t.Generic[T]):
    _inner: t.Union[T, t.List[T]]
    _is_val: bool

    def __init__(self, val: t.Union[T, t.List[T]], _is_val: bool):
        self._inner = val
        self._is_val = _is_val

    @classmethod
    def from_val(cls, val: T) -> ValueOrList[T]:
        return cls(val, True)

    @classmethod
    def from_list(cls, l: t.List[T]) -> ValueOrList[T]:
        return cls(l, False)

    def __repr__(self) -> str:
        return f"ValueOrList({self._inner!r})"

    def __str__(self) -> str:
        return str(self._inner)

    def __eq__(self, other: t.Any) -> bool:
        if not self.__class__ == other.__class__:
            return False
        return self._is_val == other._is_val and self._inner == other._inner

    @classmethod
    def _converter(cls: t.Type[T], *args: t.Type[Convertible]) -> ValueOrListConverter:
        arg = args[0] if len(args) > 0 else t.Any
        return ValueOrListConverter(arg)

    def __len__(self) -> int:
        return 1 if self._is_val else len(t.cast(t.List[T], self._inner))

    def map(self, f: t.Callable[[T], U]) -> ValueOrList[U]:
        if self._is_val:
            return ValueOrList(f(t.cast(T, self._inner)), True)
        return ValueOrList(list(map(f, t.cast(t.List[T], self._inner))), False)

    def __iter__(self) -> t.Iterator[T]:
        if self._is_val:
            yield t.cast(T, self._inner)
        else:
            yield from t.cast(t.List[T], self._inner)


class ValueOrListConverter(UnionConverter):
    def __init__(self, ty: t.Type[Convertible]):
        types = t.cast(t.Sequence[t.Type[Convertible]], (ty, t.List[ty]))
        super().__init__(types, lambda v, i: ValueOrList(v, i == 0))
        self.ty = ty

    def expected(self, plural: bool = False) -> str:
        inner = self.converters[0].expected(plural)
        return f"{inner} or sequence of {inner}"

    def into_data(self, val: t.Any) -> DataType:
        if not isinstance(val, ValueOrList):
            return into_data(val)
        return t.cast(ValueOrList[t.Any], val).map(
            lambda v: into_data(v, self.ty)
        )._inner


class YAMLDocList(list):  # type: ignore
    """
    `list` subclass representing a list of objects from YAML documents.
    """
    ...


__all__ = [
    'PositiveInt', 'NonNegativeInt', 'NegativeInt', 'NonPositiveInt',
    'PositiveFloat', 'NonNegativeFloat', 'NegativeFloat', 'NonPositiveFloat', 'FiniteFloat',
]



import pytest

from pane.field import _split_field_name, rename_field


@pytest.mark.parametrize(('field', 'out'), [
    ('test', ('test',)),
    ('test_snake_Case', ('test', 'snake', 'Case'),),
    ('test-KEBAB-Case', ('test', 'KEBAB', 'Case'),),
    ('TestPascalCase', ('Test', 'Pascal', 'Case'),),
    ('testCamelCase', ('test', 'Camel', 'Case'),),
    ('TEST_SCREAM_CASE', ('TEST', 'SCREAM', 'CASE'),),
    # conservative for now
    ('__test__', ValueError("Unable to interpret field '__test__' for automatic rename")),
])
def test_split_field(field, out):
    if isinstance(out, ValueError):
        with pytest.raises(ValueError, match=out.args[0]):
            _split_field_name(field)
    else:
        assert _split_field_name(field) == out


@pytest.mark.parametrize(('field', 'style', 'out'), [
    ('toPascalCase', 'pascal', 'ToPascalCase'),
    ('To_CAMEL_case', 'camel', 'toCamelCase'),
    ('ToScream_Case', 'scream', 'TO_SCREAM_CASE'),
    ('To_kebab_Case', 'kebab', 'to-kebab-case'),
    ('ToSnakeCase', 'snake', 'to_snake_case'),
])
def test_rename_field(field, style, out):
    assert rename_field(field, style) == out


import collections.abc
import typing as t

from typing_extensions import ParamSpec
import pytest

from pane.util import replace_typevars, collect_typevars, flatten_union_args, get_type_hints
from pane.util import pluralize, list_phrase, remove_article


T = t.TypeVar('T')
U = t.TypeVar('U')
K = t.TypeVar('K')
V = t.TypeVar('V')
P = ParamSpec('P')


@pytest.mark.parametrize(('input', 'output'), [
    (t.Dict[K, V], dict[int, V]),
    (t.Union[int, K], int),  # deduplicate union and extract single-member union
    (t.Union[int, U], t.Union[int, str]),  # flatten union and deduplicate, preserving order
    (t.Tuple[T, ...], tuple[t.Tuple[int, ...], ...]),
    (t.Tuple[T, U, K], tuple[t.Tuple[int, ...], t.Union[int, str], int]),
    (t.Callable[[int, T], U], collections.abc.Callable[[int, t.Tuple[int, ...]], t.Union[int, str]]),
    (t.Literal['test'], t.Literal['test']),
])
def test_replace_typevars(input, output):
    replacements: t.Dict[t.Union[t.TypeVar, t.ParamSpec], type] = {
        K: int,
        T: t.Tuple[int, ...],
        U: t.Union[int, str]
    }
    assert replace_typevars(input, replacements) == output


@pytest.mark.parametrize(('input', 'output'), [
    ((T, t.Callable[P, T]), (T, P)),
    (t.Tuple[int, float, T], (T,))
])
def test_collect_typevars(input, output):
    assert collect_typevars(input) == output


class TestClass:
    x: int = 5
    y: float
    z: 'str'
    w: t.Annotated[int, ()]


@pytest.mark.parametrize(('cls', 'types'), [
    (TestClass, {'x': int, 'y': float, 'z': str, 'w': t.Annotated[int, ()]}),
])
def test_get_type_hints(cls, types):
    assert types == get_type_hints(cls)


@pytest.mark.parametrize(('input', 'output'), [
    ((int, float, str), (int, float, str)),
    ((t.Union[int, float], t.Union[float, str], T), (int, float, float, str, T)),
])
def test_flatten_union_args(input, output):
    assert tuple(flatten_union_args(input)) == output


@pytest.mark.parametrize(('input', 'conj', 'output'), [
    (('word',), None, 'word'),
    (('a', 'b'), 'xor', 'a xor b'),
    (('foo', 'bar', 'baz'), None, 'foo, bar, or baz'),
    (('foo', 'bar', 'baz'), 'and', 'foo, bar, and baz'),
])
def test_list_phrase(input, conj, output):
    if conj is not None:
        assert list_phrase(input, conj) == output
    else:
        assert list_phrase(input) == output


@pytest.mark.parametrize(('input', 'output'), [
    ('the a word', 'a word'),
    ('an article', 'article'),
    ('  the break', 'break'),
    ('article', 'article'),
])
def test_remove_article(input, output):
    assert remove_article(input) == output




import sys
from pathlib import Path
from io import TextIOBase, IOBase, TextIOWrapper, BufferedIOBase
from contextlib import AbstractContextManager, nullcontext
import functools
from itertools import zip_longest
import operator
import typing as t
from threading import RLock
from typing_extensions import ParamSpec


T = t.TypeVar('T')
FileOrPath = t.Union[str, Path, TextIOBase, t.TextIO]
P = ParamSpec('P')


# mock KW_ONLY on python <3.10
try:
    from dataclasses import KW_ONLY
except ImportError:
    class _KW_ONLY_TYPE:
        pass
    KW_ONLY = _KW_ONLY_TYPE()


def _validate_file(f: t.Union[t.IO[t.AnyStr], IOBase], mode: t.Union[t.Literal['r'], t.Literal['w']]):
    if f.closed:
        raise IOError("Error: Provided file is closed.")

    if mode == 'r':
        if not f.readable():
            raise IOError("Error: Provided file not readable.")
    elif mode == 'w':
        if not f.writable():
            raise IOError("Error: Provided file not writable.")


def open_file(f: FileOrPath,
              mode: t.Literal['r', 'w'] = 'r',
              newline: t.Optional[str] = None,
              encoding: t.Optional[str] = 'utf-8') -> AbstractContextManager[TextIOBase]:
    """
    Open the given file for text I/O.

    If given a path-like, opens it with the specified settings.
    Otherwise, make an effort to reconfigure the encoding, and
    check that it is readable/writable as specified.

    Parameters:
      f: File to open/reconfigure
      mode: Mode file should be opened in
      newline: Newline mode file should be opened in
      encoding: Encoding file should be opened in
    """
    if not isinstance(f, (IOBase, t.BinaryIO, t.TextIO)):
        return open(f, mode, newline=newline, encoding=encoding)

    if isinstance(f, TextIOWrapper):
        f.reconfigure(newline=newline, encoding=encoding)
    elif isinstance(f, t.TextIO):
        f = TextIOWrapper(f.buffer, newline=newline, encoding=encoding)
    elif isinstance(f, (BufferedIOBase, t.BinaryIO)):
        f = TextIOWrapper(t.cast(t.BinaryIO, f), newline=newline, encoding=encoding)

    _validate_file(f, mode)
    return nullcontext(f)  # don't close a f we didn't open


def partition(f: t.Callable[[T], bool], iter: t.Iterable[T]) -> t.Tuple[t.Tuple[T, ...], t.Tuple[T, ...]]:
    """Partition `iter` into values that satisfy `f` and those which don't."""
    true: t.List[T] = []
    false: t.List[T] = []
    for val in iter:
        if f(val):
            true.append(val)
        else:
            false.append(val)
    return (tuple(true), tuple(false))


def pluralize(word: str, plural: t.Union[bool, int], suffix: str = 's', article: t.Optional[str] = None) -> str:
    """Pluralize `word` based on the value of `plural`."""
    if not isinstance(plural, bool):
        plural = plural != 1
    article = article + " " if article is not None and len(article) else ""
    return (word + suffix) if plural else (article + word)


def list_phrase(words: t.Sequence[str], conj: str = 'or') -> str:
    """
    Form an english list phrase from `words`, using the conjunction `conj`.
    """
    if len(words) <= 2:
        return f" {conj} ".join(words)
    return ", ".join(words[:-1]) + f", {conj} {words[-1]}"


def remove_article(s: str) -> str:
    """Remove an article from `s`, if present."""
    s = s.lstrip()
    for article in ('a ', 'an ', 'the '):
        if s.startswith(article):
            return s[len(article):]
    return s


def _collect_typevars(d: t.Dict[t.Union[t.TypeVar, ParamSpec], None], ty: t.Any):
    if isinstance(ty, type):
        pass
    elif isinstance(ty, (tuple, t.Sequence)):
        ty = t.cast(t.Sequence[t.Any], ty)
        for arg in ty:
            _collect_typevars(d, arg)
    elif hasattr(ty, '__typing_subst__') or isinstance(ty, (t.TypeVar, ParamSpec)):
        d.setdefault(ty)
    else:
        for ty in getattr(ty, '__parameters__', ()):
            d.setdefault(ty)


def collect_typevars(args: t.Any) -> t.Tuple[t.Union[t.TypeVar, ParamSpec], ...]:
    """
    Collect a list of type variables in `args`

    Preserves order but removes duplicates (i.e. type variables are returned
    in the order they are encountered, but no type variable is returned twice).

    Loosely based on `typing._collect_parameters`.
    """
    d: t.Dict[t.Union[t.TypeVar, ParamSpec], None] = {}  # relies on dicts preserving insertion order
    _collect_typevars(d, args)
    return tuple(d)


def type_union(types: t.Iterable[type]) -> type:
    return functools.reduce(operator.or_, types)


def flatten_union_args(types: t.Iterable[T]) -> t.Iterator[T]:
    """Flatten nested unions, returning a single sequence of possible union types."""
    for ty in types:
        if t.get_origin(ty) is t.Union:
            yield from flatten_union_args(t.get_args(ty))
        else:
            yield ty


def replace_typevars(ty: t.Any,
                     replacements: t.Mapping[t.Union[t.TypeVar, ParamSpec], type]) -> t.Any:
    """
    Apply a list of type-variable replacements to `ty`, and return the modified type.
    """
    if isinstance(ty, (t.TypeVar, ParamSpec)):
        return replacements.get(ty, ty)
    if isinstance(ty, t.Sequence) and not isinstance(ty, (str, bytes)):
        return type(ty)(replace_typevars(t, replacements) for t in ty)  # type: ignore

    base = t.get_origin(ty) or ty
    args = t.get_args(ty)

    if not len(args):
        return ty

    args = (replace_typevars(ty, replacements) for ty in args)

    if base is t.Union:
        args = tuple(flatten_union_args(args))
        # deduplicate union
        args = dict.fromkeys(args).keys()

        if len(args) == 1:
            # single-element union, return as value
            return next(iter(args))

    return base[tuple(args)]  # type: ignore


def get_type_hints(cls: type) -> t.Dict[str, t.Any]:
    """
    Extract a dict of type hints from `cls`. Evaluate forward refs if possible.

    This is a slightly modified version of [typing.get_type_hints]().
    """

    globalns = getattr(sys.modules.get(cls.__module__, None), '__dict__', {})
    localns = dict(vars(cls))

    d: t.Dict[str, t.Any] = {}
    for name, value in cls.__dict__.get('__annotations__', {}).items():
        if value is None:
            value = type(None)
        if isinstance(value, str):
            value = t.ForwardRef(value, is_argument=False, is_class=True)
        if isinstance(value, t.ForwardRef):
            # hack to handle top-level KW_ONLY
            val = value.__forward_value__ if value.__forward_evaluated__ else eval(value.__forward_code__, globalns, localns)
            if val is KW_ONLY:
                d[name] = KW_ONLY
                continue
        # private access inside typing module
        value = t._eval_type(value, globalns, localns)  # type: ignore
        d[name] = value

    return d


def broadcast_shapes(*args: t.Sequence[int]) -> t.Tuple[int, ...]:
    """
    Attempt to broadcast the given shapes together using numpy semantics.

    Defers to `numpy.broadcast_shapes` if numpy is available.
    """
    try:
        import numpy
        return numpy.broadcast_shapes(*map(tuple, args))
    except ImportError:
        pass

    # our own implementation, with worse error messages
    out_shape: t.List[int] = []
    for ax_lens in zip_longest(*(reversed(arg) for arg in args), fillvalue=1):
        bcast = max(ax_lens)
        if not all(ax_len in (1, bcast) for ax_len in ax_lens):
            shapes = [f"'{tuple(arg)!r}'" for arg in args]
            raise ValueError(f"Couldn't broadcast shapes {list_phrase(shapes, 'and')}")
        out_shape.append(bcast)
    return tuple(out_shape)


def is_broadcastable(*args: t.Sequence[int]) -> bool:
    """Return whether `args` are broadcastable together using numpy semantics."""
    try:
        broadcast_shapes(*args)
        return True
    except ValueError:
        return False


PREV, NEXT, KEY, RESULT = 0, 1, 2, 3

class KeyCache(t.Generic[P, T]):
    _missing = object()

    def __init__(self, f: t.Callable[P, T], key_f: t.Callable[[t.Any], t.Any] = id, maxsize: t.Optional[int] = None):
        self.maxsize: t.Optional[int] = maxsize
        self.key_f = key_f
        self.inner_f: t.Callable[P, T] = f
        self.cache: t.Dict[t.Tuple[t.Tuple[t.Any, ...], t.Tuple[t.Tuple[str, t.Any], ...]], t.Any] = {}

        self._root: t.List[t.Any] = []
        self._root[:] = [self._root, self._root, None, None]
        self._lock = RLock()

        self.full = self.maxsize == 0

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self.maxsize is None:
            key = (tuple(map(self.key_f, args)), tuple((k, self.key_f(v)) for (k, v) in kwargs.items()))
            result = self.cache.get(key, self._missing)
            if result is not self._missing:
                return t.cast(T, result)
            result = self.inner_f(*args, **kwargs)
            self.cache[key] = result
            return result

        key = (tuple(map(self.key_f, args)), tuple((k, self.key_f(v)) for (k, v) in kwargs.items()))
        with self._lock:
            link = self.cache.get(key, None)
            if link is not None:
                # extract this link
                prev_link, next_link, _key, result = link
                prev_link[NEXT] = next_link
                next_link[PREV] = prev_link

                # and move it to the end of the list
                last = self._root[PREV]
                last[NEXT] = self._root[PREV] = link
                link[PREV] = last
                link[NEXT] = self._root
                return t.cast(T, result)

        result = self.inner_f(*args, **kwargs)
        with self._lock:
            if key in self.cache:
                pass
            elif self.full:
                # turn the oldest link into the new root
                # and reuse oldroot on the end of the list
                oldroot = self._root
                oldroot[KEY] = key
                oldroot[RESULT] = result

                self._root = oldroot[NEXT]
                oldkey = self._root[KEY]
                oldresult = self._root[RESULT]  # type: ignore (we want to keep this around for a bit)
                self._root[KEY] = self._root[RESULT] = None
                del self.cache[oldkey]
                self.cache[key] = oldroot
            else:
                last = self._root[PREV]
                link = [last, self._root, key, result]
                last[NEXT] = self._root[PREV] = self.cache[key] = link
                self.full = (self.maxsize is not None and len(self.cache) >= self.maxsize)
        return result


# TODO support maxsize
def key_cache(key_f: t.Callable[[t.Any], t.Any] = id, *, maxsize: t.Optional[int] = None) -> t.Callable[[t.Callable[P, T]], KeyCache[P, T]]:
    def inner(f: t.Callable[P, T]) -> KeyCache[P, T]:
        return t.cast(KeyCache[P, T], functools.update_wrapper(KeyCache(f, key_f, maxsize), f))

    return inner


__all__ = [
    'open_file', 'list_phrase', 'pluralize', 'remove_article',
    'flatten_union_args', 'collect_typevars', 'replace_typevars', 'get_type_hints',
    'KW_ONLY',
]



"""
Annotations supported by ``pane.convert()`` and dataclasses.
"""

from __future__ import annotations

import abc
from dataclasses import dataclass
import math
import typing as t

from .util import flatten_union_args, is_broadcastable
from .util import list_phrase, pluralize, remove_article

if t.TYPE_CHECKING:
    from .convert import IntoConverter
    from .converters import Converter, ConditionalConverter


class ConvertAnnotation(abc.ABC, t.Hashable):
    """
    Abstract annotation supported by `pane`.
    """
    @abc.abstractmethod
    def _converter(self, inner_type: t.Union[Converter[t.Any], IntoConverter]) -> Converter[t.Any]:
        ...


@dataclass(frozen=True)
class Tagged(ConvertAnnotation):
    tag: str
    """Name of tag. This name will be searched in every Union member"""
    external: t.Union[bool, t.Tuple[str, str]] = False
    """
    Tagged unions can be stored three ways:
     - Internally tagged (`external=False`, default). In this format, the tags are stored inside of each object: `{tag_name: tag_value, **obj}`
     - Externally tagged (`external=True`). In this format, the tag is stored as a key outside the rest of the object: `{tag_value: obj}`
     - Adjacently tagged (`external=(tag_key, value_key)`). In this format, the tag and value are stored under separate items: `{tag_key: tag_value, value_key: obj}`

    This specification affects conversion into and out symmetrically.
    """

    # for some reason this isn't auto-generated on python 3.9
    def __hash__(self):
        return hash((self.__class__.__name__, self.tag, self.external))

    def _converter(self, inner_type: t.Union[Converter[t.Any], IntoConverter]) -> Converter[t.Any]:
        from .converters import TaggedUnionConverter
        origin = t.get_origin(inner_type)
        if origin is not t.Union:
            raise TypeError(f"'Tagged' must surround a 'Union' type.")
        types = tuple(flatten_union_args(t.get_args(inner_type)))
        return TaggedUnionConverter(types, tag=self.tag, external=self.external)


@dataclass(frozen=True)
class Condition(ConvertAnnotation):
    f: t.Callable[[t.Any], bool]
    """
    Condition/predicate function.

    This is called with a parsed value, and should return `True` if it passes the condition.
    """
    name: t.Optional[str] = None
    """Human-readable name of this condition"""
    make_expected: t.Optional[t.Callable[[str, bool], str]] = None
    """
    Given an inner `expected` string, and a boolean indicating plurality, this should return a
    formatted `expected` string including the condition.

    This is a low-level function that can be overrided for better error messages.
    """

    # for some reason this isn't auto-generated on python 3.9
    def __hash__(self):
        return hash((
            self.__class__.__name__, self.f, self.name,
        ))

    def __and__(self, other: Condition) -> Condition:
        return Condition.all(self, other)

    def __or__(self, other: Condition) -> Condition:
        return Condition.any(self, other)

    def __invert__(self) -> Condition:
        return Condition(
            lambda val: not self.f(val),
            f"not {self.cond_name()}"
        )

    @staticmethod
    def all(*conditions: Condition, make_expected: t.Optional[t.Callable[[str, bool], str]] = None) -> Condition:
        """
        Create a condition by `and`ing together multiple conditions.

        Parameters:
          conditions: Conditions to combine
          make_expected: If specified, override `make_expected` on the result `Condition`.
        """
        return Condition(
            lambda val: all(cond.f(val) for cond in conditions),
            list_phrase(tuple(cond.cond_name() for cond in conditions), 'and'),
            make_expected
        )

    @staticmethod
    def any(*conditions: Condition, make_expected: t.Optional[t.Callable[[str, bool], str]] = None) -> Condition:
        """
        Create a condition by `or`ing together multiple conditions.

        Parameters:
          conditions: Conditions to combine
          make_expected: If specified, override `make_expected` on the result `Condition`.
        """
        return Condition(
            lambda val: any(cond.f(val) for cond in conditions),
            list_phrase(tuple(cond.cond_name() for cond in conditions), 'or'),
            make_expected
        )

    def cond_name(self) -> str:
        """Get the name of this condition"""
        return self.name or self.f.__name__

    def _converter(self, inner_type: t.Union[Converter[t.Any], IntoConverter]) -> ConditionalConverter[t.Any]:
        from .converters import ConditionalConverter
        return ConditionalConverter(
            inner_type, self.f, self.cond_name(),
            self.make_expected or (lambda conv, plural: f"{conv} satisfying {self.cond_name()}"),
        )


def val_range(*, min: t.Union[int, float, None] = None, max: t.Union[int, float, None] = None) -> Condition:
    """`Condition` indicating that a value must be between `min` and `max` (inclusive)."""
    conds: t.List[Condition] = []
    if min is not None:
        conds.append(Condition(lambda v: v >= min, f"v >= {min}"))
    if max is not None:
        conds.append(Condition(lambda v: v <= max, f"v <= {max}"))
    return Condition.all(*conds)


def len_range(*, min: t.Optional[int] = None, max: t.Optional[int] = None) -> Condition:
    """`Condition` indicating that a value must have between `min` and `max` elements (inclusive)."""
    conds: t.List[Condition] = []
    if min is not None:
        conds.append(Condition(lambda v: len(v) >= min, f"at least {min} {pluralize('elem', min)}"))
    if max is not None:
        conds.append(Condition(lambda v: len(v) <= max, f"at most {max} {pluralize('elem', max)}"))
    cond = Condition.all(*conds, make_expected=lambda exp, plural: f"{exp} with {cond.cond_name()}")
    return cond


def shape(shape: t.Sequence[int]) -> Condition:
    """
    `Condition` indicating that a value must have a shape `shape`.

    Fails on objects that don't have a `shape` attribute.
    """
    name = f"shape {tuple(shape)}"
    return Condition(
        lambda v: v.shape == shape, name,
        lambda exp, plural: f"{exp} with {name}"
    )


def broadcastable(shape: t.Sequence[int]) -> Condition:
    """
    `Condition` indicating that a value must be broadcastable to shape `shape`.

    Fails on objects that don't have a `shape` attribute.
    """
    name = f"broadcastable to {tuple(shape)}"
    return Condition(
        lambda v: is_broadcastable(v.shape, shape), name,
        lambda exp, plural: f"{exp} {name}"
    )


def adjective_condition(f: t.Callable[[t.Any], bool], adjective: str, article: str = 'a') -> Condition:
    """
    Make a condition that can be expressed as a simple adjective (e.g. 'empty' or 'non-empty').

    Parameters:
      f: Condition/predicate function
      adjective: Adjective corresponding to `f` (e.g. 'empty' or 'non-empty').
      article: Article to put in front of `adjective` (only when not pluralized).
    """
    return Condition(
        f, adjective,
        # e.g. 'positive ints' if plural else 'a positive int'
        lambda exp, plural: f"{adjective} {exp}" if plural else f"{article} {adjective} {remove_article(exp)}"
    )


Positive = adjective_condition(lambda v: v > 0, 'positive')
"""`Condition` indicating value must be positive"""
Negative = adjective_condition(lambda v: v < 0, 'negative')
"""`Condition` indicating value must be negative"""
NonPositive = adjective_condition(lambda v: v <= 0, 'non-positive')
"""`Condition` indicating value must be non-positive"""
NonNegative = adjective_condition(lambda v: v >= 0, 'non-negative')
"""`Condition` indicating value must be non-negative"""
Finite = adjective_condition(math.isfinite, 'finite')
"""`Condition` indicating value must be finite"""
Empty = adjective_condition(lambda v: len(v) == 0, 'empty')
"""`Condition` indicating value must be empty (have no elements)"""
NonEmpty = adjective_condition(lambda v: len(v) != 0, 'non-empty')
"""`Condition` indicating value must not be empty"""


__all__ = [
    'ConvertAnnotation', 'Tagged', 'Condition', 'val_range', 'len_range',
    'Positive', 'Negative', 'NonPositive', 'NonNegative',
    'Finite', 'Empty', 'NonEmpty',
]



from __future__ import annotations

import inspect
import typing as t

import pytest

import pane
from pane.errors import ErrorNode, ProductErrorNode, DuplicateKeyError, WrongTypeError
from pane.annotations import Tagged


def check_ord(obj, other, ordering: t.Literal[-1, 0, 1]):
    assert (obj < other) == (ordering < 0)
    assert (obj > other) == (ordering > 0)
    assert (obj <= other) == (ordering <= 0)
    assert (obj >= other) == (ordering >= 0)
    assert (obj == other) == (ordering == 0)
    assert (obj != other) == (ordering != 0)


def f(*args, **kwargs):
    return (args, kwargs)


def test_pane_required_after_default():
    with pytest.raises(TypeError, match="Mandatory field 'y' follows optional field"):
        class RequiredAfterDefault(pane.PaneBase):
            x: int = 3
            y: int

    class RequiredAfterDefaultKwOnly(pane.PaneBase):
        _: pane.KW_ONLY
        x: int = 3
        y: int


def test_pane_required_kw_only():
    class RequiredKwOnly(pane.PaneBase):
        x: int
        z: int = pane.field(kw_only=True)
        y: int

    with pytest.raises(TypeError, match="Field 'z' is kw_only but mandatory. This is incompatible with the 'tuple' in_format."):
        class RequiredKwOnlyTuple(RequiredKwOnly, in_format='tuple'):
            ...


class TestClass(pane.PaneBase, out_format='tuple'):
    x: int = 3
    y: float = 5.

    __test__ = False


@pytest.mark.parametrize(('args', 'result'), [
    (f(), {'x': 3, 'y': 5.}),
    (f(4), {'x': 4, 'y': 5.}),
    (f(y=8), {'x': 3, 'y': 8.0}),
])
def test_pane_construction(args, result):
    (args, kwargs) = args
    assert TestClass(*args, **kwargs).dict() == result


@pytest.mark.parametrize(('args', 'result', 'set_only_result'), [
    (f(), {'x': 3, 'y': 5.}, {}),
    (f('st'), {'x': 'st', 'y': 5.}, {'x': 'st'}),
    (f(y=None), {'x': 3, 'y': None}, {'y': None}),
])
def test_pane_unchecked(args, result, set_only_result):
    (args, kwargs) = args
    assert TestClass.make_unchecked(*args, **kwargs).dict() == result
    assert TestClass.make_unchecked(*args, **kwargs).dict(set_only=True) == set_only_result


@pytest.mark.parametrize(('args', 'result'), [
    (f(), "TestClass(x=3, y=5.0)"),
    (f(4), "TestClass(x=4, y=5.0)"),
    (f(y=8), "TestClass(x=3, y=8.0)"),
])
def test_pane_repr(args, result):
    (args, kwargs) = args
    assert repr(TestClass(*args, **kwargs)) == result


def test_pane_ord():
    check_ord(TestClass(x=1, y=3.), TestClass(x=2, y=1.), -1)
    check_ord(TestClass(x=2, y=3.), TestClass(x=2, y=1.), 1)
    check_ord(TestClass(x=2, y=1.), TestClass(x=2, y=1.), 0)
    with pytest.raises(TypeError):
        check_ord(TestClass(x=1, y=3.), 5, -1)


class TestClassInherit(TestClass, out_format='struct'):
    z: float = pane.field(default=3., kw_only=True)


class TestClassInherit2(TestClassInherit):
    w: int = 4


class TestClassModifyFields(TestClassInherit):
    x: float = 9.  # type: ignore


class TestClass2(pane.PaneBase, in_format=('tuple', 'struct')):
    x: int = 1
    z: int = pane.field(default=3, kw_only=True)
    y: int = 2
    _: pane.KW_ONLY
    w: int = pane.field(default=4, aliases=('W', 'P'))

    __test__ = False


class TestClass3(pane.PaneBase, kw_only=True, allow_extra=True):
    x: int = 1
    z: int = pane.field(default=3, kw_only=True)
    y: int = 2
    _: pane.KW_ONLY
    w: int

    __test__ = False


@pytest.mark.parametrize(('cls', 'sig'), [
    (TestClass, '(x: int = 3, y: float = 5.0) -> None'),
    (TestClass2, '(x: int = 1, y: int = 2, *, z: int = 3, w: int = 4) -> None'),
    (TestClass3, '(*, x: int = 1, z: int = 3, y: int = 2, w: int) -> None'),
    (TestClassInherit, '(x: int = 3, y: float = 5.0, *, z: float = 3.0) -> None'),
    (TestClassInherit2, '(x: int = 3, y: float = 5.0, w: int = 4, *, z: float = 3.0) -> None'),
    (TestClassModifyFields, '(x: float = 9.0, y: float = 5.0, *, z: float = 3.0) -> None'),
])
def test_init_signature(cls, sig):
    assert str(inspect.signature(cls.__init__)) == sig
    assert str(inspect.signature(cls)) == sig


@pytest.mark.parametrize(('cls', 'sig'), [
    (TestClass, '(x: int = 3, y: float = 5.0) -> test_pane.TestClass'),
    (TestClass2, '(x: int = 1, y: int = 2, *, z: int = 3, w: int = 4) -> test_pane.TestClass2'),
    (TestClass3, '(*, x: int = 1, z: int = 3, y: int = 2, w: int) -> test_pane.TestClass3'),
])
def test_make_unchecked_signature(cls, sig):
    assert str(inspect.signature(cls.make_unchecked)) == sig


@pytest.mark.parametrize(('cls', 'val', 'result'), [
    (TestClass, {'x': 3, 'y': 5.}, TestClass(3, 5.)),
    # extra fields
    (TestClass, {'x': 3, 'y': 5., 'zz': 5}, ProductErrorNode('struct TestClass', {}, {'x': 3, 'y': 5., 'zz': 5}, extra={'zz'})),
    # duplicate keys
    (TestClass2, {'x': 3, 'w': 3, 'W': 4}, ProductErrorNode('struct TestClass2', {'W': DuplicateKeyError('W', ('w', 'W', 'P'))}, {'x': 3, 'w': 3, 'W': 4})),
    # extra fields, allow_extra=True
    (TestClass3, {'x': 2, 'y': 3, 'w': 4, 'zz': 5}, TestClass3(x=2, y=3, w=4)),
    # missing fields
    (TestClass3, {}, ProductErrorNode('struct TestClass3', {}, {}, missing={'w'}))
])
def test_pane_convert(cls, val, result):
    if isinstance(result, ErrorNode):
        with pytest.raises(pane.ConvertError) as e:
            pane.convert(val, cls)
        assert e.value.tree == result
    else:
        assert pane.convert(val, cls) == result


@pytest.mark.parametrize(('val', 'result'), [
    # out_format tuple
    (TestClass(x=3, y=5.), (3, 5.)),
    # out_format struct
    (TestClass2(x=3, w=3), {'x': 3, 'y': 2, 'w': 3, 'z': 3}),
])
def test_pane_into_data(val: pane.PaneBase, result: pane.DataType):
    assert val.into_data() == result


T = t.TypeVar('T')
U = t.TypeVar('U')


class GenericPane(pane.PaneBase, t.Generic[T]):
    x: T

class GenericInherit(GenericPane[U]):
    y: U

class ConcreteGenericInherit(GenericPane[int]):
    y: float

class GenericGenericInherit(GenericInherit[T]):
    z: T

class GenericConcreteInherit(GenericInherit[int], t.Generic[T]):
    z: T


@pytest.mark.parametrize(('lhs', 'rhs', 'result'), [
    (GenericPane[int](5), GenericPane[t.Any](5), True),
    (TestClass3(x=1, z=3, w=3), TestClass3(w=3), True),
    (TestClass(), TestClass2(), False),
])
def test_pane_eq(lhs, rhs, result):
    if result:
        assert lhs == rhs
    else:
        assert lhs != rhs


@pytest.mark.parametrize(('cls', 'sig'), [
    (GenericPane, '(x: ~T) -> None'),
    (GenericPane[int], '(x: int) -> None'),
    (ConcreteGenericInherit, '(x: int, y: float) -> None'),
    (GenericInherit, '(x: ~U, y: ~U) -> None'),
    (GenericInherit[float], '(x: float, y: float) -> None'),
    (GenericGenericInherit, '(x: ~T, y: ~T, z: ~T) -> None'),
    (GenericGenericInherit[float], '(x: float, y: float, z: float) -> None'),
    (GenericConcreteInherit, '(x: int, y: int, z: ~T) -> None'),
    (GenericConcreteInherit[float], '(x: int, y: int, z: float) -> None'),
])
def test_generic_signatures(cls, sig):
    assert str(inspect.signature(cls.__init__)) == sig
    assert str(inspect.signature(cls)) == sig


@pytest.mark.parametrize(('cls', 'args', 'error'), [
    (GenericPane[int], f('s'), WrongTypeError('an int', 's')),
    (GenericPane[int], f(5), None),
    (GenericPane, f('any value'), UserWarning("Unbound TypeVar '~T'. Will be interpreted as 'typing.Any'.")),
    (GenericInherit[float], f('s', 5.), WrongTypeError('a float', 's')),
    (GenericInherit[float], f(5., 's'), WrongTypeError('a float', 's')),
])
def test_generic_pane_init(cls, args, error):
    (args, kwargs) = args

    if isinstance(error, ErrorNode):
        with pytest.raises(pane.ConvertError) as e:
            cls(*args, **kwargs)
        assert e.value.tree == error
    elif isinstance(error, Warning):
        with pytest.warns(type(error), match=error.args[0]):
            cls(*args, **kwargs)
    else:
        cls(*args, **kwargs)


class PaneRename(pane.PaneBase):
    snake_case: int
    SCREAM_CASE: float
    camelCase: int
    PascalCase: float

class PaneRenameKebab(PaneRename, rename='kebab'):
    ...


def test_pane_rename():
    obj = {'snake-case': 1, 'scream-case': 2., 'camel-case': 3, 'pascal-case': 4.}
    assert PaneRenameKebab.from_data(obj) == PaneRenameKebab(1, 2., 3, 4.)
    assert PaneRenameKebab(1, 2., 3, 4.).into_data() == obj


class PaneTag1(pane.PaneBase, kw_only=True):
    tag: t.Literal['tag1'] = 'tag1'
    x: int
    y: float


class PaneTag2(pane.PaneBase, kw_only=True):
    tag: t.Literal['tag2'] = 'tag2'
    x: int


class PaneTag3(pane.PaneBase, kw_only=True):
    tag: t.Literal['tag3'] = 'tag3'
    x: int


class PaneTagBase(pane.PaneBase):
    x: t.Annotated[t.Union[PaneTag1, PaneTag2, PaneTag3], Tagged('tag')]


@pytest.mark.parametrize(('val', 'result'), [
    ({'x': {'tag': 'tag1', 'x': 5, 'y': 5.0}}, PaneTagBase(x=PaneTag1(x=5, y=5.0))),
    ({'x': {'tag': 'tag3', 'x': 8}}, PaneTagBase(x=PaneTag3(x=8))),
])
def test_pane_convert_tagged_union(val, result):
    cls = PaneTagBase
    if isinstance(result, ErrorNode):
        with pytest.raises(pane.ConvertError) as e:
            pane.convert(val, cls)
        assert e.value.tree == result
    else:
        assert pane.convert(val, cls) == result


def test_manual_slots():
    class SlotsClass(pane.PaneBase, frozen=False):
        __slots__ = ("x", "z")
        x: int

    obj = SlotsClass(5)
    obj.z = 8
    with pytest.raises(AttributeError, match="'SlotsClass' object has no attribute 'y'"):
        obj.y = 8


"""
Converter types, which do the hard work of recursive validation.
"""

# pyright: reportUnknownMemberType=none

import abc
import dataclasses
import re
import traceback
import datetime
import enum
from fractions import Fraction
from decimal import Decimal
import typing as t
from typing_extensions import TypeGuard

from pane.errors import ErrorNode

from .convert import DataType, Convertible, IntoConverter, make_converter, into_data
from .util import list_phrase, pluralize, flatten_union_args, type_union, KW_ONLY
from .errors import ConvertError, ParseInterrupt, WrongTypeError, ConditionFailedError
from .errors import ErrorNode, SumErrorNode, ProductErrorNode


T_co = t.TypeVar('T_co', covariant=True)
T = t.TypeVar('T')
U = t.TypeVar('U', bound=Convertible)
FromDataT = t.TypeVar('FromDataT', bound=Convertible)
FromDataK = t.TypeVar('FromDataK', bound=Convertible)
FromDataV = t.TypeVar('FromDataV', bound=Convertible)
NestedSequence = t.Union[T, t.Sequence['NestedSequence[T]']]
DatetimeT = t.TypeVar('DatetimeT', bound=t.Union[datetime.datetime, datetime.date, datetime.time])


def data_is_sequence(val: t.Any) -> TypeGuard[t.Sequence[t.Any]]:
    """Return whether `val` is a sequence-like data type."""
    return isinstance(val, t.Sequence) and not isinstance(val, (str, bytes, bytearray))


def data_is_mapping(val: t.Any) -> TypeGuard[t.Mapping[t.Any, t.Any]]:
    """Return whether `val` is a mapping-like data type."""
    return isinstance(val, (dict, t.Mapping))


class Converter(abc.ABC, t.Generic[T_co]):
    """
    Base class for a converter to a given type ``T_co``.
    """

    def convert(self, val: t.Any) -> T_co:
        """Convert ``val`` to ``T_co``. Raises a ``ConvertError`` on failure."""
        try:
            return self.try_convert(val)
        except ParseInterrupt:
            pass
        node = self.collect_errors(val)
        if node is None:
            raise RuntimeError("convert() raised but ``collect_errors`` returned ``None``."
                               " This is a bug of the ``Converter`` implementation.")
        raise ConvertError(node)

    def into_data(self, val: t.Any) -> DataType:
        """
        Convert ``val`` into a data interchange format.

        ``val`` *should* be of a type returned by this converter,
        but don't count on it.
        """
        return into_data(val, None)

    @abc.abstractmethod
    def expected(self, plural: bool = False) -> str:
        """
        Return a descriptive string indicating the value(s) expected.

        Parameters:
          plural: Whether to pluralize the descriptive string
        """
        ...

    @abc.abstractmethod
    def try_convert(self, val: t.Any) -> T_co:
        """
        Attempt to convert ``val`` to ``T``.
        Should raise ``ParseInterrupt`` (and only ``ParseInterrupt``)
        when a given parsing path reaches a dead end.
        """
        ...

    @abc.abstractmethod
    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """
        Return an error tree caused by converting ``val`` to ``T``.
        ``collect_errors`` should return ``None`` iff ``convert`` doesn't raise.
        """
        ...


@dataclasses.dataclass
class AnyConverter(Converter[t.Any]):
    """Converter for ``t.Any``."""
    def try_convert(self, val: t.Any) -> t.Any:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        return val

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return pluralize("any value", plural)

    def collect_errors(self, val: t.Any) -> None:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        return None


@dataclasses.dataclass
class ScalarConverter(Converter[T]):
    """
    Converter for a simple scalar type,
    constructible from a list of allowed types.
    """

    # TODO this needs to handle into_data better
    ty: t.Type[T]
    """Type to convert into."""
    allowed: t.Union[type, t.Tuple[type, ...]]
    """Type or list of allowed types."""
    expect: t.Optional[str] = None
    """Singular form of expected value."""
    expect_plural: t.Optional[str] = None
    """Plural form of expected value."""
    _into_data_f: t.Callable[[T], DataType] = lambda v: v  # type: ignore

    def __post_init__(self):
        self.expect = self.expect or self.ty.__name__
        self.expect_plural = self.expect_plural or self.expect

    def into_data(self, val: t.Any) -> DataType:
        return self._into_data_f(val)

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return t.cast(str, self.expect_plural if plural else self.expect)

    def try_convert(self, val: t.Any) -> T:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if isinstance(val, self.allowed):
            try:
                return self.ty(val)  # type: ignore
            except Exception:
                raise ParseInterrupt()
        raise ParseInterrupt()

    def collect_errors(self, val: t.Any) -> t.Optional[WrongTypeError]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if isinstance(val, self.allowed):
            try:
                self.ty(val)
                return None
            except Exception as e:
                tb = e.__traceback__.tb_next  # type: ignore
                tb = traceback.TracebackException(type(e), e, tb)
                return WrongTypeError(self.expected(), val, tb)
        return WrongTypeError(f'{self.expected()}', val)


@dataclasses.dataclass
class NoneConverter(Converter[None]):
    """
    Converter which accepts only ``None``.
    """

    def try_convert(self, val: t.Any) -> None:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if val is None:
            return val
        raise ParseInterrupt()

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return pluralize("null value", plural)

    def collect_errors(self, val: t.Any) -> t.Optional[WrongTypeError]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if val is None:
            return None
        return WrongTypeError(self.expected(), val)


@dataclasses.dataclass
class LiteralConverter(Converter[T_co]):
    """
    Converter which accepts any of a list of literal values.
    """

    vals: t.Sequence[T_co]

    def try_convert(self, val: t.Any) -> T_co:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if val in self.vals:
            return val
        raise ParseInterrupt()

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        l = list_phrase(tuple(map(repr, self.vals)))
        return f"({l})" if plural else l

    def collect_errors(self, val: t.Any) -> t.Optional[WrongTypeError]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if val in self.vals:
            return None
        return WrongTypeError(self.expected(), val)


@dataclasses.dataclass(init=False)
class UnionConverter(Converter[t.Any]):
    """
    Converter for an untagged union of subtypes.
    Unions are always evaluated left-to-right.
    """
    types: t.Tuple[IntoConverter, ...]
    """List of potential types"""
    converters: t.Tuple[Converter[t.Any], ...]
    """List of type converters"""
    constructor: t.Optional[t.Callable[[t.Any, int], t.Any]]
    """
    Constructor to call with parsed value.
    Called with ``(val, index of type in union)``
    """

    def __init__(self, types: t.Sequence[IntoConverter],
                 constructor: t.Optional[t.Callable[[t.Any, int], t.Any]] = None):
        self.types = tuple(flatten_union_args(types))
        self.converters = tuple(map(make_converter, types))
        self.constructor = constructor

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return list_phrase(tuple(conv.expected(plural) for conv in self.converters))

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        # this is tricky, because we have no type information about which variant ``val`` is.
        # so we basically try_convert each until we find a match
        # this works because try_convert should be idempotent
        for conv in self.converters:
            try:
                conv.try_convert(val)
            except ParseInterrupt:
                pass
            else:
                return conv.into_data(val)
        # default to regular conversion
        return into_data(val)

    def construct(self, val: t.Any, i: int) -> t.Any:
        if self.constructor is None:
            return val
        return self.constructor(val, i)

    def try_convert(self, val: t.Any) -> t.Any:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        for (i, conv) in enumerate(self.converters):
            try:
                val = conv.try_convert(val)
                try:
                    return self.construct(val, i)
                except Exception:
                    pass
            except ParseInterrupt:
                pass
        raise ParseInterrupt

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        failed_children: t.List[ErrorNode] = []
        for (i, conv) in enumerate(self.converters):
            # if one branch is successful, the whole type is successful
            try:
                conv_val = conv.try_convert(val)
            except ParseInterrupt:
                failed_children.append(t.cast(t.Union[ProductErrorNode, WrongTypeError], conv.collect_errors(val)))
                continue
            try:
                self.construct(conv_val, i)
                return None
            except Exception as e:
                tb = e.__traceback__.tb_next  # type: ignore
                tb = traceback.TracebackException(type(e), e, tb)
                failed_children.append(WrongTypeError(self.expected(), val, tb))
        return SumErrorNode(failed_children)


@dataclasses.dataclass(init=False)
class TaggedUnionConverter(UnionConverter):
    """
    Converter for a tagged union of subtypes.

    Tagged unions may be parsed in three ways. The default
    is an 'internally tagged' union, where the tag is found
    by looking for a given key in the given object. This is the
    default.
    An 'externally tagged' union is stored as a dict with a key
    and a single value ``{tag: content}``. This may be specified
    using ``external=True``.
    Finally, a 'adjacently tagged' union may be specified using
    two keys ``external=(t, c)``. The union is stored as
    ``{t: tag, c: content}``
    """
    tag: str
    tag_map: t.Dict[t.Any, int]
    """Map from tags to indices into self.types/self.converters"""
    external: t.Union[bool, t.Tuple[str, str]] = False
    """
    Tagged union representation.
    False: internal representation
    True: external representation
    (t, c): adjacent representation
    """

    def __init__(self, types: t.Sequence[t.Any], tag: str,
                 external: t.Union[bool, t.Tuple[str, str]] = False):
        super().__init__(types)
        self.tag = tag
        self.external = external if isinstance(external, t.Sequence) else bool(external)

        # look for tag in each of self.types
        self.tag_map = {}
        for (i, ty) in enumerate(self.types):
            try:
                # TODO error if used on non-literal
                val = getattr(ty, self.tag)
                if val in self.tag_map:
                    raise TypeError(f"Tag value '{val}' matches multiple types")
                self.tag_map[val] = i
            except AttributeError:
                raise AttributeError(f"Tag '{self.tag}' not found inside type '{ty}'")

    def tag_expected(self) -> str:
        """Return a string list of the expected/supported tags"""
        return list_phrase(tuple(map(repr, self.tag_map.keys())))

    def obj_expected(self, plural: bool = False) -> str:
        """Return a string list of the supported objects"""
        return list_phrase(tuple(conv.expected(plural) for conv in self.converters))

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        if self.external is False:
            # internally tagged
            return self.obj_expected(plural)
        obj = self.obj_expected(False)
        mapping = pluralize('mapping', plural, article='a')
        tag = list_phrase(tuple(map(repr, self.tag_map.keys())))
        if self.external is True:
            # externally tagged
            return f"{mapping} '{tag}' => {obj}"
        else:
            # adjacently tagged
            (t, c) = self.external 
            return f"{mapping} {repr(t)} => {tag}, {repr(c)} => {obj}"

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        tag = getattr(val, self.tag)
        inner_conv = self.converters[self.tag_map[tag]]
        if self.external is False:
            # internally tagged
            return inner_conv.into_data(val)
        if self.external is True:
            # externally tagged
            return {tag: inner_conv.into_data(val)}
        # adjacently tagged
        (t_r, c_r) = self.external
        return {t_r: tag, c_r: inner_conv.into_data(val)}

    def try_convert(self, val: t.Any) -> t.Any:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if not data_is_mapping(val):
            raise ParseInterrupt()
        val = t.cast(t.Dict[str, t.Any], val)
        tag: t.Any

        if self.external is False:
            try:
                # don't give 'tag' to variants
                val = val.copy()
                tag = val.pop(self.tag)
            except KeyError:
                raise ParseInterrupt()
        elif self.external is True:
            if len(val) != 1:
                raise ParseInterrupt()
            (tag, val) = next(iter(val.items()))
        else:
            (t_r, c_r) = self.external
            try:
                if len(val) != 2:
                    raise ParseInterrupt()
                tag, val = val[t_r], val[c_r]
            except KeyError:
                raise ParseInterrupt()
        try:
            i = self.tag_map[tag]
        except KeyError:
            raise ParseInterrupt()
        return self.converters[i].try_convert(val)

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if not data_is_mapping(val):
            return WrongTypeError(self.expected(), val)
        val = t.cast(t.Dict[str, t.Any], val)
        tag: t.Any

        if self.external is False:
            try:
                # don't give 'tag' to variants
                val = val.copy()
                tag = val.pop(self.tag)
            except KeyError:
                return WrongTypeError(f"mapping with key '{self.tag}' => {self.tag_expected()}", val)
        elif self.external is True:
            if len(val) != 1:
                return WrongTypeError(self.expected(), val)
            (tag, val) = next(iter(val.items()))
        else:
            (t_r, c_r) = self.external
            try:
                if len(val) != 2:
                    raise KeyError()
                tag, val = val[t_r], val[c_r]
            except KeyError:
                return WrongTypeError(f"mapping with keys '{t_r}' and '{c_r}'", val)
        try:
            i = self.tag_map[tag]
        except KeyError:
            return WrongTypeError(f"tag '{self.tag}' one of {self.tag_expected()}", tag)
        return self.converters[i].collect_errors(val)


@dataclasses.dataclass
class StructConverter(Converter[T]):
    """
    Converter for a simple, hetereogeneous struct-like type, constructible from a dict.
    """

    ty: t.Type[T]
    """Type to convert into. Must be constructible from a dict/mapping."""
    fields: t.Mapping[str, IntoConverter]
    """List of fields and their types"""
    name: t.Optional[str] = None
    """Optional name of struct"""

    _: KW_ONLY = dataclasses.field(init=False, repr=False, compare=False)
    opt_fields: t.Set[str] = dataclasses.field(default_factory=set)
    """Set of fields which are optional"""
    field_converters: t.Dict[str, Converter[t.Any]] = dataclasses.field(init=False)
    """Dict of sub-converters for each field"""

    def __post_init__(self):
        self.field_converters = {k: make_converter(v) for (k, v) in self.fields.items()}

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        name = f" {self.name}" if self.name is not None else ""
        return f"{pluralize('struct', plural)}{name}"

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        assert data_is_mapping(val)
        d: t.Dict[DataType, DataType] = {}
        for (k, v) in t.cast(t.Mapping[str, t.Any], val).items():
            if (conv := self.field_converters.get(k)) is not None:
                d[k] = conv.into_data(v)
            else:
                d[k] = into_data(v)
        return d

    def try_convert(self, val: t.Any) -> T:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if not data_is_mapping(val):
            raise ParseInterrupt()
        val = t.cast(t.Dict[str, t.Any], val)
        d: t.Dict[str, t.Any] = {}
        for (k, v) in val.items():
            if k not in self.fields:
                raise ParseInterrupt()  # unknown field
            d[k] = self.field_converters[k].try_convert(v)
        missing = set(self.fields.keys()) - set(val.keys()) - self.opt_fields
        if len(missing):
            raise ParseInterrupt()
        return self.ty(d)  # type: ignore

    def collect_errors(self, val: t.Any) -> t.Union[WrongTypeError, ProductErrorNode, None]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if not data_is_mapping(val):
            return WrongTypeError(self.expected(), val)
        val = t.cast(t.Dict[str, t.Any], val)

        children: t.Dict[t.Union[str, int], t.Any] = {}
        extra: t.Set[str] = set()
        for (k, v) in val.items():
            if k not in self.fields:
                extra.add(k)
                continue
            if (node := self.field_converters[k].collect_errors(v)) is not None:
                children[k] = node
        missing = set(self.fields.keys()) - set(val.keys()) - self.opt_fields
        if len(children) or len(missing) or len(extra):
            return ProductErrorNode(self.expected(), children, val, missing, extra)
        return None


@dataclasses.dataclass(init=False)
class TupleConverter(t.Generic[T], Converter[T]):
    """Converter for a simple, heterogeneous tuple-like type"""
    ty: t.Type[T]
    """Type to convert into. Must be constructible from a sequence/tuple"""
    converters: t.Tuple[Converter[t.Any], ...]
    """List of sub-converters for each field"""

    def __init__(self, ty: t.Type[T], types: t.Sequence[IntoConverter]):
        self.ty = ty
        self.converters = tuple(map(make_converter, types))

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        return tuple(
            conv.into_data(v)
            for (v, conv) in zip(t.cast(t.Sequence[t.Any], val), self.converters)
        )

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return f"{pluralize('tuple', plural)} of length {len(self.converters)}"

    def try_convert(self, val: t.Any) -> T:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if not data_is_sequence(val):
            raise ParseInterrupt
        if len(val) != len(self.converters):
            raise ParseInterrupt

        return self.ty(conv.try_convert(v) for (conv, v) in zip(self.converters, val))

    def collect_errors(self, val: t.Any) -> t.Union[None, ProductErrorNode, WrongTypeError]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if not data_is_sequence(val) or len(val) != len(self.converters):
            return WrongTypeError(self.expected(), val)
        children = {}
        for (i, (conv, v)) in enumerate(zip(self.converters, val)):
            node = conv.collect_errors(v)
            if node is not None:
                children[i] = node
        if len(children) == 0:
            return None
        return ProductErrorNode(self.expected(), children, val)


@dataclasses.dataclass(init=False)
class DictConverter(t.Generic[FromDataK, FromDataV], Converter[t.Mapping[FromDataK, FromDataV]]):
    """Converter for a homogenous dict-like type."""
    ty: t.Type[t.Mapping[FromDataK, FromDataV]]
    """Type to convert into. Must be constructible from a dict (unless `constructor` is specified)"""
    k_conv: Converter[FromDataK]
    """Sub-converter for keys"""
    v_conv: Converter[FromDataV]
    """Sub-converter for values"""
    constructor: t.Callable[[t.Dict[t.Any, t.Any]], t.Mapping[FromDataK, FromDataV]]

    def __init__(self, ty: t.Type[t.Dict[t.Any, t.Any]],
                 k: t.Type[FromDataK] = t.Any, v: t.Type[FromDataV] = t.Any,
                 constructor: t.Optional[t.Callable[[t.Dict[t.Any, t.Any]], t.Mapping[FromDataK, FromDataV]]] = None):
        self.ty = ty
        self.k_conv = make_converter(k)
        self.v_conv = make_converter(v)
        self.constructor = self.ty if constructor is None else constructor

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        return {
            self.k_conv.into_data(k): self.v_conv.into_data(v)
            for (k, v) in t.cast(t.Mapping[FromDataK, FromDataV], val).items()
        }

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return f"{pluralize('mapping', plural)} of {self.k_conv.expected(True)} => {self.v_conv.expected(True)}"

    def try_convert(self, val: t.Any) -> t.Mapping[FromDataK, FromDataV]:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if not data_is_mapping(val):
            raise ParseInterrupt()

        d = {self.k_conv.try_convert(k): self.v_conv.try_convert(v) for (k, v) in val.items()}
        # TODO catch errors here
        return self.constructor(d)

    def collect_errors(self, val: t.Any) -> t.Union[None, WrongTypeError, ProductErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if not data_is_mapping(val):
            return WrongTypeError(self.expected(), val)

        nodes: t.Dict[t.Union[str, int], ErrorNode] = {}
        for (k, v) in val.items():
            if (node := self.k_conv.collect_errors(k)) is not None:
                nodes[str(k)] = node  # TODO split bad fields from bad values
            if (node := self.v_conv.collect_errors(v)) is not None:
                nodes[str(k)] = node
        if len(nodes):
            return ProductErrorNode(self.expected(), nodes, val)


@dataclasses.dataclass(init=False)
class SequenceConverter(t.Generic[FromDataT], Converter[t.Sequence[FromDataT]]):
    """Converter for a homogenous sequence-like type"""
    ty: type
    """Type to convert into. Must be constructible from an iterator."""
    v_conv: Converter[FromDataT]
    """Sub-converter for values"""
    constructor: t.Callable[[t.Iterator[t.Any]], t.Sequence[t.Any]]

    def __init__(self, ty: t.Type[t.Sequence[t.Any]], v: t.Type[FromDataT] = t.Any,
                 constructor: t.Optional[t.Callable[[t.Iterator[t.Any]], t.Sequence[t.Any]]] = None):
        self.ty = ty
        self.v_conv = make_converter(v)
        self.constructor = self.ty if constructor is None else constructor

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        return [
            self.v_conv.into_data(v)
            for v in t.cast(t.Sequence[FromDataT], val)
        ]

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return f"{pluralize('sequence', plural)} of {self.v_conv.expected(True)}"

    def try_convert(self, val: t.Any) -> t.Sequence[FromDataT]:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if not data_is_sequence(val):
            raise ParseInterrupt
        try:
            return self.constructor(self.v_conv.try_convert(v) for v in val)  # type: ignore
        except Exception:
            raise ParseInterrupt()

    def collect_errors(self, val: t.Any) -> t.Union[None, WrongTypeError, ProductErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if not data_is_sequence(val):
            return WrongTypeError(self.expected(), val)

        nodes = {}
        vals: t.List[FromDataT] = []
        for (i, v) in enumerate(val):
            try:
                vals.append(self.v_conv.convert(v))
            except ConvertError as e:
                nodes[i] = e.tree

        if len(nodes):
            return ProductErrorNode(self.expected(), nodes, val)
        # try to construct val
        try:
            self.constructor(iter(vals))
            return None
        except Exception as e:
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return WrongTypeError(self.expected(), val, tb)


@dataclasses.dataclass
class NestedSequenceConverter(t.Generic[T, U], Converter[T]):
    """
    Converter which delegates to a sub-converter, and then attempts
    to construct a different type
    """
    val_type: t.Type[U]
    """Inner type to convert to"""
    constructor: t.Callable[[NestedSequence[U]], T]
    """Constructor to call."""
    ragged: bool = False
    """Whether to accept ragged arrays."""
    val_conv: Converter[U] = dataclasses.field(init=False)
    """[`Converter`][pane.converters.Converter] for value type"""

    def __post_init__(self):
        self.val_conv = make_converter(self.val_type)

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        word = 'nested sequence' if self.ragged else 'n-d array'
        return f"{pluralize(word, plural, article='a')} of {self.val_conv.expected(True)}"

    @staticmethod
    def _check_shape(val: NestedSequence[t.Any], dim: int = 0) -> t.Tuple[int, ...]:
        if not data_is_sequence(val):
            # single value
            return ()
        shapes = [NestedSequenceConverter._check_shape(v, dim+1) for v in val]
        if len(shapes) == 0:
            return (0,)
        shape = shapes[0]
        if not all(s == shape for s in shapes):
            raise ValueError(f"shape mismatch at dim {dim}. Sub-shapes: {shapes}")
        new_shape = (len(shapes), *shape)
        return new_shape

    def try_convert(self, val: t.Any) -> T:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        result = self._try_convert(val)
        if not self.ragged:
            try:
                self._check_shape(result)
            except ValueError:
                raise ParseInterrupt()
        return self.constructor(result)

    def _try_convert(self, val: t.Any) -> NestedSequence[U]:
        if not data_is_sequence(val):
            # single value
            return self.val_conv.try_convert(val)
        vals = list(map(self._try_convert, val))
        return t.cast(NestedSequence[U], vals)

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if (node := self._collect_errors(val)) is not None:
            return node
        val = self._try_convert(val)
        if not self.ragged:
            try:
                self._check_shape(val)
            except ValueError as e:
                return WrongTypeError(self.expected(), val, info=e.args[0])
        try:
            self.constructor(val)
        except Exception as e:
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return WrongTypeError(self.expected(), val, tb)

    def _collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        if not data_is_sequence(val):
            return self.val_conv.collect_errors(val)
        nodes = {}
        for (i, v) in enumerate(val):
            if (node := self._collect_errors(v)) is not None:
                nodes[i] = node
        if len(nodes):
            return ProductErrorNode(self.expected(), nodes, val)


@dataclasses.dataclass
class ConditionalConverter(t.Generic[FromDataT], Converter[FromDataT]):
    """
    Converter which applies an arbitrary pre-condition to the converted value.
    """
    inner_type: t.Union[t.Type[FromDataT], Converter[FromDataT]]
    """Inner type to apply condition to"""
    condition: t.Callable[[FromDataT], bool]
    """Function to evaluate condition"""
    condition_name: str
    """Human-readable name of condition"""
    make_expected: t.Callable[[str, bool], str]
    """Function which takes ``(expected, plural)`` and makes a compound ``expected``."""
    inner: Converter[FromDataT] = dataclasses.field(init=False)
    """Inner sub-converter"""

    def __post_init__(self):
        if isinstance(self.inner_type, Converter):
            self.inner = self.inner_type
        else:
            self.inner = make_converter(t.cast(t.Type[FromDataT], self.inner_type))

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        return self.inner.into_data(val)

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return self.make_expected(self.inner.expected(plural), plural)

    def try_convert(self, val: t.Any) -> FromDataT:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        val = self.inner.try_convert(val)
        try:
            if self.condition(val):
                return val
        except Exception:
            pass
        raise ParseInterrupt()

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        try:
            conv_val = self.inner.try_convert(val)
        except ParseInterrupt:
            # TODO with_expected() here
            return self.inner.collect_errors(val)
        try:
            # condition failed
            if not self.condition(conv_val):
                return ConditionFailedError(self.expected(), val, self.condition_name)
        except Exception as e:
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return ConditionFailedError(self.expected(), val, self.condition_name, tb)
        return None


class EnumConverter(Converter[enum.Enum]):
    def __init__(self, ty: t.Type[enum.Enum]):
        from pane.convert import _DataType  # type: ignore
        if issubclass(ty, enum.Flag):
            raise TypeError("Flag enums are not currently supported")
        self.ty: t.Type[enum.Enum] = ty

        members = ty.__members__.values()
        try:
            self.val_map = {member.value: member for member in members}
        except TypeError:
            raise TypeError("All enum members must be hashable")

        self.member_vals = tuple(self.val_map.keys())
        if not all(isinstance(val, _DataType) for val in self.member_vals):
            raise TypeError("All enum members must be data-interchange types")

        self.inner_ty = type_union(map(type, self.member_vals))
        self.inner_conv = make_converter(self.inner_ty)

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        if isinstance(val, self.ty):
            return val.value  # guaranteed to be data-interchange type
        return into_data(val)

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        l = list_phrase(tuple(map(str, self.member_vals)))
        return f"{pluralize('member', plural)} of enum '{self.ty.__name__}' ({l})"

    def try_convert(self, val: t.Any) -> enum.Enum:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        val = self.inner_conv.try_convert(val)
        try:
            return self.val_map[val]
        except KeyError:
            raise ParseInterrupt()

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        try:
            val = self.inner_conv.try_convert(val)
        except ParseInterrupt:
            return self.inner_conv.collect_errors(val)
        try:
            self.val_map[val]
            return None
        except KeyError:
            return WrongTypeError(self.expected(), val)


@dataclasses.dataclass
class DelegateConverter(t.Generic[T, U], Converter[T]):
    """
    Converter which delegates to a sub-converter, and then attempts
    to construct a different type
    """
    from_type: t.Type[U]
    """Inner type to convert to"""
    constructor: t.Callable[[U], T]
    """Constructor for outer type"""
    expect: t.Optional[str] = None
    """Expected value. Defaults to inner expected value."""
    expect_plural: t.Optional[str] = None
    """Plural expected value. Defaults to inner expected value."""
    inner: Converter[U] = dataclasses.field(init=False)
    """Inner sub-converter"""

    def __post_init__(self):
        self.inner = make_converter(self.from_type)
        self.expect = self.expect or self.from_type.__name__
        self.expect_plural = self.expect_plural or self.expect

    def into_data(self, val: t.Any) -> DataType:
        """See [`Converter.into_data`][pane.converters.Converter.into_data]"""
        # TODO: this is a hack, because we can't easily convert T back to U
        try:
            return self.inner.into_data(val)
        except Exception:
            pass
        return into_data(val)

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return t.cast(str, self.expect_plural if plural else self.expect)

    def try_convert(self, val: t.Any) -> T:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        val = self.inner.try_convert(val)
        try:
            return self.constructor(val)
        except Exception:
            raise ParseInterrupt from None

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        try:
            conv_val = self.inner.try_convert(val)
        except ParseInterrupt:
            # TODO with_expected() here
            return self.inner.collect_errors(val)
        try:
            self.constructor(conv_val)
        except Exception as e:
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return WrongTypeError(self.expected(), val, tb)


@dataclasses.dataclass(init=False)
class PatternConverter(t.Generic[t.AnyStr], Converter[re.Pattern[t.AnyStr]]):
    ty: t.Type[t.AnyStr]
    ty_conv: Converter[t.AnyStr]

    def __init__(self, ty: t.Type[t.AnyStr] = str, *args: t.Any):
        if len(args) > 0:
            raise TypeError("PatternConverter takes only one type argument")
        self.ty = ty
        if not issubclass(ty, (str, bytes)):
            raise TypeError(f"Pattern only accepts a 'str' or 'bytes' type argument, instead got '{ty!r}'")
        self.ty_conv = make_converter(self.ty)

    def into_data(self, val: t.Any) -> t.AnyStr:
        assert isinstance(val, re.Pattern)
        return t.cast(re.Pattern[t.AnyStr], val).pattern

    def expected(self, plural: bool = False) -> str:
        ty = 'bytes' if self.ty is bytes else 'string'
        return pluralize(f'{ty} regex pattern', plural, article='a')

    def try_convert(self, val: t.Any) -> re.Pattern[t.AnyStr]:
        if isinstance(val, re.Pattern):
            val = t.cast(re.Pattern[t.Any], val).pattern
        s = self.ty_conv.try_convert(val)
        try:
            return re.compile(s)
        except Exception:
            raise ParseInterrupt from None

    def collect_errors(self, val: t.Any) -> t.Optional[ErrorNode]:
        if isinstance(val, re.Pattern):
            val = t.cast(re.Pattern[t.Any], val).pattern
        try:
            s = self.ty_conv.try_convert(val)
        except ParseInterrupt:
            return WrongTypeError(self.expected(), val)
        try:
            re.compile(s)
        except re.error as e:
            tb = e.__traceback__.tb_next  # type: ignore
            tb = traceback.TracebackException(type(e), e, tb)
            return WrongTypeError(self.expected(), val, tb)
        return None


class DatetimeConverter(Converter[DatetimeT], t.Generic[DatetimeT]):
    """
    Converter for a simple scalar type,
    constructible from a list of allowed types.
    """
    _date_types: t.Tuple[type, ...] = (datetime.date, datetime.time, datetime.datetime)
    _expected: t.Mapping[type, str] = {
        datetime.date: "date",
        datetime.datetime: "datetime",
        datetime.time: "time",
    }

    def __init__(self, ty: t.Type[DatetimeT]):
        self.ty = ty
        self.super_ty: t.Type[DatetimeT]
        if ty in self._date_types:
            self.super_ty = ty
            return
        for date_ty in self._date_types:
            if issubclass(ty, date_ty):
                self.super_ty = t.cast(t.Type[DatetimeT], date_ty)
                return
        raise TypeError(f"Only types {list_phrase([repr(str(ty)) for ty in self._date_types])} are supported")

    def __eq__(self, other: t.Any) -> bool:
        return type(self) == type(other) and self.ty == other.ty

    def expected(self, plural: bool = False) -> str:
        """See [`Converter.expected`][pane.converters.Converter.expected]"""
        return pluralize(self._expected[self.super_ty], plural, article='a')

    def into_data(self, val: t.Any) -> str:
        if isinstance(val, (datetime.time, datetime.date, datetime.datetime)):
            return val.isoformat()
        return str(val)

    #                input type:
    # output type:     date    datetime  time    str
    #          date     id     .date()   error  parse
    #      datetime   combine     id     error  parse
    #          time    error   .timetz()  id    parse

    def from_datetime(self, dt: datetime.datetime) -> DatetimeT:
        d: t.Mapping[type, t.Callable[[datetime.datetime], DatetimeT]] = {
            # datetime to datetime
            datetime.datetime: lambda dt: t.cast(DatetimeT, dt),
            # datetime to time
            datetime.time: lambda dt: t.cast(DatetimeT, dt.time()),
            # datetime to date
            datetime.date: lambda dt: t.cast(DatetimeT, dt.date()),
        }
        return d[self.super_ty](dt)

    def from_date(self, dt: datetime.date) -> DatetimeT:
        def err(val: t.Any):
            raise TypeError()

        d: t.Mapping[type, t.Callable[[datetime.date], DatetimeT]] = {
            # date to datetime
            datetime.datetime: lambda date: t.cast(DatetimeT, datetime.datetime.combine(date, datetime.time())),
            # date to date
            datetime.date: lambda date: t.cast(DatetimeT, date),
            datetime.time: err,
        }
        return d[self.super_ty](dt)

    def try_convert(self, val: t.Any) -> DatetimeT:
        """See [`Converter.try_convert`][pane.converters.Converter.try_convert]"""
        if isinstance(val, str):
            # parse string
            try:
                return t.cast(DatetimeT, self.ty.fromisoformat(val))
            except ValueError:
                raise ParseInterrupt() from None
        if isinstance(val, datetime.datetime):
            # from datetime, to datetime, date, or time
            return self.from_datetime(val)
        elif isinstance(val, datetime.time):
            # from time, to time only
            if self.super_ty == datetime.time:
                return t.cast(DatetimeT, val)
        elif isinstance(val, datetime.date):
            # from date, to date or datetime
            if self.super_ty != datetime.time:
                return self.from_date(val)
        raise ParseInterrupt()

    def collect_errors(self, val: t.Any) -> t.Optional[WrongTypeError]:
        """See [`Converter.collect_errors`][pane.converters.Converter.collect_errors]"""
        if isinstance(val, str):
            # parse string
            try:
                self.ty.fromisoformat(val)
                return None
            except ValueError as e:
                tb = e.__traceback__.tb_next  # type: ignore
                tb = traceback.TracebackException(type(e), e, tb)
                return WrongTypeError(self.expected(), val, tb)
        if isinstance(val, datetime.datetime):
            return None
        elif isinstance(val, datetime.time):
            if self.super_ty == datetime.time:
                return None
        elif isinstance(val, datetime.date):
            if self.super_ty != datetime.time:
                return None
        return WrongTypeError(self.expected(), val)


# converters for scalar types
_BASIC_CONVERTERS: t.Dict[type, Converter[t.Any]] = {
    complex: ScalarConverter(complex, (int, float, complex), 'a complex float', 'complex floats'),
    float: ScalarConverter(float, (int, float), 'a float', 'floats'),
    int: ScalarConverter(int, int, 'an int', 'ints'),
    str: ScalarConverter(str, str, 'a string', 'strings'),
    bytes: ScalarConverter(bytes, (bytes, bytearray), 'a bytestring', 'bytestrings'),
    bytearray: ScalarConverter(bytearray, (bytes, bytearray), 'a bytearray', 'bytearrays'),
    type(None): NoneConverter(),
    datetime.datetime: DatetimeConverter(datetime.datetime),
    datetime.time: DatetimeConverter(datetime.time),
    datetime.date: DatetimeConverter(datetime.date),
    Decimal: ScalarConverter(Decimal, (int, str, float, Decimal), 'a decimal number', 'decimal numbers', str),
    Fraction: ScalarConverter(Fraction, (int, str, float, Decimal, Fraction), 'a fraction', 'fractions', str),
}
"""Built-in scalar converters for some basic types"""

_BASIC_WITH_ARGS: t.Dict[type, t.Type[Converter[t.Any]]] = {
    re.Pattern: PatternConverter,
    t.Pattern: PatternConverter,
}

__all__ = [
    'Converter', 'AnyConverter', 'ScalarConverter', 'NoneConverter', 'LiteralConverter',
    'UnionConverter', 'StructConverter', 'TupleConverter', 'DictConverter', 'SequenceConverter',
    'DelegateConverter', 'ConditionalConverter',
]
